""" Event Module """

import inspect


class Event:
    def __init__(self, application):
        """Event contructor
        Arguments:
            application - The Masonite application class
        """
        self.application = application
        self.events = {}

    def get_events(self):
        return self.events

    def listen(self, event, listeners):

        if not isinstance(listeners, list):
            listeners = [listeners]

        if event in self.events:
            self.events[event] += listeners
        else:
            self.events.update({event: listeners})

        return self

    def fire(self, event, *args, **kwargs):
        if isinstance(event, str):
            collected_events = self.collect_events(event)
            for collected_event in collected_events:
                for listener in self.events.get(collected_event, []):
                    listener().handle(event, *args, **kwargs)
            return collected_events
        else:
            if inspect.isclass(event):
                lookup = event
                event = event()
            else:
                lookup = event.__class__
            for listener in self.events.get(lookup, []):
                listener().handle(event, *args, **kwargs)

            return [event]

    def collect_events(self, fired_event):
        collected_events = []
        for stored_event in self.events.keys():

            if not isinstance(stored_event, str):
                continue

            if stored_event == fired_event:
                collected_events.append(fired_event)

            elif stored_event.endswith("*") and fired_event.startswith(
                stored_event.replace("*", "")
            ):
                collected_events.append(stored_event)

            elif stored_event.startswith("*") and fired_event.endswith(
                stored_event.replace("*", "")
            ):
                collected_events.append(stored_event)

            elif "*" in stored_event:
                starts, end = stored_event.split("*")
                if fired_event.startswith(starts) and fired_event.endswith(end):
                    collected_events.append(stored_event)

        return collected_events

    def subscribe(self, *listeners):
        """Subscribe a specific listener object to the events system
        Raises:
            InvalidSubscriptionType -- raises when the subscribe attribute on the listener object is not a class.
        """
        for listener in listeners:
            listener.subscribe(self)



from tests import TestCase
from src.masonite.notification import SlackMessage


class WelcomeToSlack(SlackMessage):
    def build(self):
        return (
            self.to("#general")
            .from_("sam")
            .text("Hello from Masonite!")
            .link_names()
            .unfurl_links()
            .without_markdown()
            .can_reply()
            .mode(2)  # API MODE
        )


class TestSlackMessage(TestCase):
    def test_build_message(self):
        slack_message = WelcomeToSlack().build().get_options()
        self.assertEqual(slack_message.get("channel"), "#general")
        self.assertEqual(slack_message.get("username"), "sam")
        self.assertEqual(slack_message.get("text"), "Hello from Masonite!")
        self.assertEqual(slack_message.get("link_names"), True)
        self.assertEqual(slack_message.get("unfurl_links"), True)
        self.assertEqual(slack_message.get("unfurl_media"), True)
        self.assertEqual(slack_message.get("mrkdwn"), False)
        self.assertEqual(slack_message.get("reply_broadcast"), True)
        self.assertEqual(slack_message.get("as_user"), False)
        self.assertEqual(slack_message.get("blocks"), "[]")

    def test_from_options(self):
        slack_message = SlackMessage().from_("sam", icon=":ghost").build().get_options()
        self.assertEqual(slack_message.get("username"), "sam")
        self.assertEqual(slack_message.get("icon_emoji"), ":ghost")
        self.assertEqual(slack_message.get("icon_url"), "")
        slack_message = SlackMessage().from_("sam", url="#").build().get_options()
        self.assertEqual(slack_message.get("username"), "sam")
        self.assertEqual(slack_message.get("icon_url"), "#")
        self.assertEqual(slack_message.get("icon_emoji"), "")

    def test_build_with_blocks(self):
        from slackblocks import DividerBlock, HeaderBlock

        slack_message = (
            SlackMessage()
            .from_("Sam")
            .text("Hello")
            .block(HeaderBlock("Header title", block_id="1"))
            .block(DividerBlock(block_id="2"))
            .build()
            .get_options()
        )
        self.assertEqual(
            slack_message.get("blocks"),
            '[{"type": "header", "block_id": "1", "text": {"type": "plain_text", "text": "Header title"}}, {"type": "divider", "block_id": "2"}]',
        )

    def test_api_mode_options(self):
        slack_message = SlackMessage().as_current_user().token("123456")
        self.assertEqual(slack_message._token, "123456")
        slack_message_options = slack_message.build().get_options()
        self.assertEqual(slack_message_options.get("as_user"), True)



import bcrypt


class BcryptHasher:
    def __init__(self, options={}):
        self.options = options

    def set_options(self, options):
        self.options = options
        return self

    def make(self, string):
        return self.make_bytes(string).decode("utf-8")

    def make_bytes(self, string):
        rounds = self.options.get("rounds", 12)
        salt = bcrypt.gensalt(rounds=rounds)
        return bcrypt.hashpw(bytes(string, "utf-8"), salt)

    def check(self, plain_string, hashed_string):
        if not isinstance(hashed_string, bytes):
            hashed_string = bytes(hashed_string or "", "utf-8")
        return bcrypt.checkpw(bytes(plain_string, "utf-8"), hashed_string)

    def needs_rehash(self, hashed_string):
        # Bcrypt hashes have the format $2b${rounds}${salt}{checksum}. rounds is encoded as
        # 2 zero-padded decimal digits. The prefix (2b) is never modified in make() function so we
        # can assume that rounds value used when generating the hash is located at [4:6] indexes
        # of the hash.
        old_rounds_value = int(hashed_string[4:6])
        return old_rounds_value != self.options.get("rounds", 12)



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..notification import Notification as NotificationObject

class Notification:
    """Notification handler facade, which handle sending/queuing notifications anonymously
    or to notifiables through different channels."""

    def add_driver(name: str, driver: str): ...
    def get_driver(name: str) -> Any: ...
    def set_configuration(config: dict) -> "Notification": ...
    def get_config_options(driver: str) -> dict: ...
    def send(
        notifiables: list,
        notification: "NotificationObject",
        drivers: list = [],
        dry: bool = False,
        fail_silently: bool = False,
    ) -> Any:
        """Send the given notification to the given notifiables."""
        ...
    def route(driver: str, route: str) -> Any:
        """Specify how to send a notification to an anonymous notifiable."""
        ...



from typing import TYPE_CHECKING
from ..validation import Validator

if TYPE_CHECKING:
    from ..validation import RuleEnclosure, MessageBag


class ValidatesRequest:
    """Request mixin to add inputs validation to requests."""

    def validate(self, *rules: "str|dict|RuleEnclosure") -> "MessageBag":
        """Validate request inputs against the given rules."""
        validator = Validator()
        return validator.validate(self.all(), *rules)



import os
from tests import TestCase

from src.masonite.utils.location import (
    base_path,
    views_path,
    controllers_path,
    seeds_path,
    migrations_path,
    config_path,
    jobs_path,
    resources_path,
    models_path,
)


class TestLocation(TestCase):
    def test_base_path(self):
        base_dir = os.getcwd()
        location = base_path()
        self.assertEqual(base_dir, location)
        location = base_path("tests/integrations")
        self.assertEqual(os.path.join(base_dir, "tests/integrations"), location)

    def test_views_path(self):
        location = views_path("app.html")
        self.assertTrue(location.endswith("tests/integrations/templates/app.html"))
        location = views_path("account/app.html")
        self.assertTrue(
            location.endswith("tests/integrations/templates/account/app.html")
        )
        location = views_path("account/app.html", absolute=False)
        self.assertEqual("tests/integrations/templates/account/app.html", location)
        location = views_path(absolute=False)
        self.assertEqual(location, "tests/integrations/templates/")

    def test_controllers_path(self):
        location = controllers_path("MyController.py")
        self.assertTrue(
            location.endswith("tests/integrations/controllers/MyController.py")
        )
        location = controllers_path("account/MyController.py")
        self.assertTrue(
            location.endswith("tests/integrations/controllers/account/MyController.py")
        )
        location = controllers_path("MyController.py", absolute=False)
        self.assertEqual("tests/integrations/controllers/MyController.py", location)

    def test_config_path(self):
        location = config_path("app.py")
        self.assertTrue(location.endswith("tests/integrations/config/app.py"))
        location = config_path("package/base.py")
        self.assertTrue(location.endswith("tests/integrations/config/package/base.py"))
        location = config_path("app.py", absolute=False)
        self.assertEqual("tests/integrations/config/app.py", location)

    def test_migrations_path(self):
        location = migrations_path("create_users_table.py")
        self.assertTrue(
            location.endswith(
                "tests/integrations/databases/migrations/create_users_table.py"
            )
        )
        location = migrations_path("package/create_team_table.py")
        self.assertTrue(
            location.endswith(
                "tests/integrations/databases/migrations/package/create_team_table.py"
            )
        )
        location = migrations_path("create_users_table.py", absolute=False)
        self.assertEqual(
            "tests/integrations/databases/migrations/create_users_table.py", location
        )

    def test_seeds_path(self):
        location = seeds_path("create_users.py")
        self.assertTrue(
            location.endswith("tests/integrations/databases/seeds/create_users.py")
        )
        location = seeds_path("package/create_teams.py")
        self.assertTrue(
            location.endswith(
                "tests/integrations/databases/seeds/package/create_teams.py"
            )
        )
        location = seeds_path("create_users.py", absolute=False)
        self.assertEqual("tests/integrations/databases/seeds/create_users.py", location)

    def test_jobs_path(self):
        location = jobs_path("SomeTask.py")
        self.assertTrue(location.endswith("tests/integrations/jobs/SomeTask.py"))
        location = jobs_path("critical/SomeTask.py")
        self.assertTrue(
            location.endswith("tests/integrations/jobs/critical/SomeTask.py")
        )
        location = jobs_path("critical/SomeTask.py", absolute=False)
        self.assertEqual("tests/integrations/jobs/critical/SomeTask.py", location)
        location = jobs_path(absolute=False)
        self.assertEqual(location, "tests/integrations/jobs/")

    def test_resources_path(self):
        location = resources_path("js/Home.vue")
        self.assertTrue(location.endswith("tests/integrations/resources/js/Home.vue"))
        location = resources_path("js/Home.vue", absolute=False)
        self.assertEqual("tests/integrations/resources/js/Home.vue", location)
        location = resources_path(absolute=False)
        self.assertEqual(location, "tests/integrations/resources/")

    def test_models_path(self):
        location = models_path("User.py")
        self.assertTrue(location.endswith("tests/integrations/app/User.py"))
        location = models_path("admin/Log.py")
        self.assertTrue(location.endswith("tests/integrations/app/admin/Log.py"))
        location = models_path("admin/Log.py", absolute=False)
        self.assertEqual("tests/integrations/app/admin/Log.py", location)
        location = models_path(absolute=False)
        self.assertEqual(location, "tests/integrations/app/")



from tests import TestCase
from src.masonite.routes import Route


class TestResponseHelpers(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.get("/test-with-errors", "WelcomeController@with_errors"),
            Route.get("/test-with-input", "WelcomeController@form_with_input"),
        )

    def test_with_input(self):
        self.get("/test-with-input", {"name": "Sam"}).assertSessionHas("name", "Sam")

    def test_with_errors(self):
        self.get("/test-with-errors").assertSessionHasErrors().assertSessionHasErrors(
            ["email"]
        )



from masonite.commands import Command


class __class__(Command):
    """
    New command description

    command-slug
        {arg1 : Argument 1 description}
        {--o|--option1 : Option 1 description}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        # do whatever you want !

        self.info("Success !")



from ..hashing import Hash
from ..hashing.drivers import BcryptHasher, Argon2Hasher
from .Provider import Provider
from ..configuration import config


class HashServiceProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        hashing = Hash(self.application).set_configuration(
            config("application.hashing")
        )
        hashing.add_driver("bcrypt", BcryptHasher())
        hashing.add_driver("argon2", Argon2Hasher())
        self.application.bind("hash", hashing)

    def boot(self):
        pass



import unittest

from src.masonite.cookies import CookieJar
from src.masonite.utils.time import cookie_expire_time


class TestCookies(unittest.TestCase):
    def test_cookies_can_get_set(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie", "name")
        self.assertEqual(cookiejar.get("cookie").value, "name")
        self.assertEqual(cookiejar.get("cookie").name, "cookie")

    def test_cookies_can_put_to_dict(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie", "name")
        self.assertEqual(cookiejar.to_dict(), {"cookie": "name"})

    def test_cookie_jar_can_render_cookie_string(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie", "name", http_only=False)
        self.assertEqual(
            cookiejar.render_response(),
            [("Set-Cookie", "cookie=name;Path=/;SameSite=Strict;")],
        )

    def test_cookie_jar_can_render_multiple_cookies(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie1", "name", http_only=False)
        cookiejar.add("cookie2", "name", http_only=False)
        self.assertEqual(
            cookiejar.render_response(),
            [
                ("Set-Cookie", "cookie1=name;Path=/;SameSite=Strict;"),
                ("Set-Cookie", "cookie2=name;Path=/;SameSite=Strict;"),
            ],
        )

    def test_cookie_jar_can_render_multiple_cookies_with_different_options(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie1", "name", path="/")
        self.assertEqual(
            cookiejar.render_response(),
            [("Set-Cookie", "cookie1=name;HttpOnly;Path=/;SameSite=Strict;")],
        )

    def test_cookie_with_expires(self):
        cookiejar = CookieJar()
        time = cookie_expire_time("2 months")
        cookiejar.add("cookie1", "name", path="/", expires=time, timezone="GMT")
        self.assertEqual(
            cookiejar.render_response(),
            [
                (
                    "Set-Cookie",
                    f"cookie1=name;HttpOnly;Expires={time} GMT;Path=/;SameSite=Strict;",
                )
            ],
        )

    def test_cookie_with_expired_already(self):
        cookiejar = CookieJar()
        time = cookie_expire_time("expired")
        cookiejar.add("cookie1", "name", path="/", expires=time, timezone="GMT")
        self.assertEqual(
            cookiejar.render_response(),
            [
                (
                    "Set-Cookie",
                    f"cookie1=name;HttpOnly;Expires={time} GMT;Path=/;SameSite=Strict;",
                )
            ],
        )

    def test_cookie_can_load(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie1", "name", http_only=False)
        cookiejar.load("csrf_token=tok")
        self.assertEqual(
            cookiejar.render_response(),
            [("Set-Cookie", "cookie1=name;Path=/;SameSite=Strict;")],
        )
        self.assertEqual(cookiejar.get("csrf_token").value, "tok")

    def test_cookie_can_make_secure_cookies(self):
        cookiejar = CookieJar()
        cookiejar.add("cookie1", "name", http_only=False, secure=True)
        self.assertEqual(
            cookiejar.render_response(),
            [("Set-Cookie", "cookie1=name;Secure;Path=/;SameSite=Strict;")],
        )



from typing import TYPE_CHECKING, Any, Tuple, List

if TYPE_CHECKING:
    from ..routes import Route

class Auth:
    """Authentication facade."""

    def add_guard(name: str, guard: Any): ...
    def set_configuration(config: dict): ...
    def guard(guard: Any) -> "Auth": ...
    def get_guard(name: str = None) -> Any: ...
    def get_config_options(guard: Any = None) -> dict: ...
    def attempt(email: str, password: str, once: bool = False) -> Any: ...
    def attempt_by_id(user_id: int, once: bool = False) -> Any: ...
    def logout(self) -> "Auth":
        """Logout the current authenticated user."""
        ...
    def user(self) -> Any:
        """Get the current authenticated user."""
        ...
    def register(dictionary: dict) -> Any:
        """"""
        ...
    def password_reset(email: str) -> "Tuple[None, None]|Tuple[int,str]":
        """Reset password of the user with the given email."""
        ...
    def reset_password(password: str, token: str) -> bool:
        """Reset password of the user with the given authentication token."""
        ...
    @classmethod
    def routes(self) -> List[Route]: ...



import pendulum
from .Cookie import Cookie

from ..utils.time import cookie_expire_time


class CookieJar:
    def __init__(self):
        self.cookies = {}
        self.loaded_cookies = {}
        self.deleted_cookies = {}

    def add(self, name, value, **options):
        self.cookies.update({name: Cookie(name, value, **options)})

    def all(self):
        cookies = self.loaded_cookies
        cookies.update(self.cookies)
        return cookies

    def all_added(self):
        return self.cookies

    def get(self, name):
        aggregate = self.all()
        return aggregate.get(name)

    def exists(self, name):
        return name in self.cookies or name in self.loaded_cookies

    def is_expired(self, name):
        cookie = self.get(name)
        return cookie.expires < pendulum.now()

    def delete(self, name):
        self.deleted_cookies.update(
            {
                name: Cookie(
                    name, "", expires=cookie_expire_time("expired"), timezone="GMT"
                )
            }
        )
        if name in self.cookies:
            self.cookies.pop(name)

        if name in self.loaded_cookies:
            self.loaded_cookies.pop(name)

    def load_cookie(self, key, value):
        self.loaded_cookies.update({key: Cookie(key, value)})

    def to_dict(self):
        dic = {}
        aggregate = self.cookies
        aggregate.update(self.loaded_cookies)
        for name, cookie in aggregate.items():
            dic.update({name: cookie.value})

        return dic

    def load(self, cookie_string):
        for compound_value in cookie_string.split("; "):
            if "=" in compound_value:
                key, value = compound_value.split("=", 1)
                self.load_cookie(key, value)
        return self

    def render_response(self):
        cookies = []
        for name, cookie in {**self.deleted_cookies, **self.all_added()}.items():
            cookies.append(("Set-Cookie", cookie.render()))

        return cookies



from ..views import View
from .Provider import Provider


class ViewProvider(Provider):
    def __init__(self, app):
        self.application = app

    def register(self):
        view = View(self.application)
        view.add_location(self.application.make("views.location"))

        self.application.bind("view", view)

    def boot(self):
        pass



from src.masonite.packages import PackageProvider
from tests.integrations.test_package.commands.Command1 import Command1
from tests.integrations.test_package.commands.Command2 import Command2
from ..TestPreset import TestPreset

"""
.../mypackage/templates/
 - admin.html
 - test.html

../myproject/templates/
"""


class MyTestPackageProvider(PackageProvider):
    def configure(self):
        (
            self.root("tests/integrations/test_package")
            .name("test_package")
            .config("config/test.py", publish=True)
            .views("templates", publish=True)
            .commands(Command1(), Command2())
            .migrations(
                "migrations/create_some_table.py", "migrations/create_other_table.py"
            )
            .assets("assets")
            .controllers("controllers")  # ensure this one is done before routes()
            .routes("routes/api.py", "routes/web.py")
            .presets(TestPreset())
        )



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..mail import Mailable

class Mail:
    """Mail facade."""

    def add_driver(name: str, driver: Any): ...
    def set_configuration(config: dict) -> "Mail": ...
    def get_driver(name: str = None) -> Any: ...
    def get_config_options(driver: str = None) -> dict: ...
    def mailable(mailable: "Mailable") -> "Mail": ...
    def send(driver: str = None): ...



from tests import TestCase
from masoniteorm.models import Model

from src.masonite.authorization import Authorizes
from src.masonite.facades import Gate


class User(Model, Authorizes):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class TestAuthorizes(TestCase):
    def setUp(self):
        super().setUp()
        self.make_request()

    def test_user_can(self):
        user = User.find(1)
        Gate.define("create-post", lambda user: user.email == "idmann509@gmail.com")
        self.assertTrue(user.can("create-post"))

    def test_user_cannot(self):
        user = User.find(1)
        Gate.define("delete-post", lambda user: False)
        self.assertTrue(user.cannot("delete-post"))
        Gate.define("view-admin-panel", lambda user: user.email == "admin@gmail.com")
        self.assertTrue(user.cannot("view-admin-panel"))



from typing import Callable, Any
from jinja2 import PackageLoader, BaseLoader

class View:
    """View facade."""

    def render(template: str, dictionary: dict = {}) -> "View":
        """Render the given template name with the given context as string."""
        ...
    def get_content(self) -> str:
        """Get the rendered content as string."""
        ...
    def hydrate_from_composers(self):
        """Add data into the view from specified composers."""
        ...
    def composer(composer_name: str, dictionary: dict) -> "View":
        """Add/Update composer with the given name and data."""
        ...
    def share(dictionary: dict) -> "View":
        """Share data to all templates."""
        ...
    def exists(template: str) -> bool:
        """Check if a template with the given name exists."""
        ...
    def add_location(template_location: str, loader: "BaseLoader" = PackageLoader):
        """Add location directory from which view templates can be loaded. The Jinja2 loader type
        can be specified."""
        ...
    def add_namespaced_location(namespace: str, template_location: str):
        """Add namespaced location directory from which view templates can be loaded."""
        ...
    def add_from_package(package_name: str, path_in_package: str): ...
    def filter(name: str, function: Callable):
        """Add filter functions to views with the given name."""
        ...
    def add_extension(extension: str) -> "View":
        """Register Jinja2 extension to views."""
        ...
    def load_template(template: str):
        """Private method for loading all the locations into the current environment."""
        ...
    def get_current_loaders(self):
        """Get all enabled Jinja2 loaders."""
        ...
    def set_separator(token: str) -> "View":
        """Change separator for view names (default is /)."""
        ...
    def set_file_extension(extension: str) -> "View":
        """Change file view extension (default is .html)."""
        ...
    def get_response(self) -> str:
        """Get the rendered content as string."""
        ...
    def test(key: str, obj: Any) -> "View": ...



from tests import TestCase


class TestValidation(TestCase):
    def test_can_validate_request(self):
        request = self.make_request(
            data={"QUERY_STRING": "email=joe@masoniteproject.com"}
        )
        validation = request.validate(
            {
                "email": "required",
            }
        )

        self.assertEqual(validation.all(), {})

    def test_can_validate_request(self):
        request = self.make_request(query_string="")
        validation = request.validate(
            {
                "email": "required",
            }
        )

        self.assertEqual(validation.all(), {"email": ["The email field is required."]})



""" A ScheduleRunCommand Command """
from cleo import Command


class ScheduleRunCommand(Command):
    """
    Run the scheduled tasks
    schedule:run
        {--t|task=? : Name of task you want to run (else all scheduled tasks will be ran)}
        {--f|force : Force running task immediately}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        return self.app.make("scheduler").run(
            run_name=self.option("task"), force=self.option("force")
        )



from src.masonite.controllers import Controller
from src.masonite.views import View
from tests.integrations.app.User import User
from src.masonite.request import Request
from src.masonite.response import Response
from src.masonite.authentication import Auth


class RegisterController(Controller):
    def show(self, view: View):  # Show register page
        return view.render("auth.register")

    def store(
        self, auth: Auth, request: Request, response: Response
    ):  # store register user
        user = auth.register(request.only("name", "email", "password"))

        if not user:
            return response.redirect("/register")

        return response.redirect("/home")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cleo import Application as CommandApplication
    from cleo import Command


class CommandCapsule:
    def __init__(self, command_application: "CommandApplication"):
        self.command_application = command_application
        self.commands = []
        self.command_name = []

    def add(self, *commands: "Command") -> "CommandCapsule":
        """Register new commands in the application."""
        for command in commands:
            command_name = command.config.name
            if command_name in self.command_name:
                continue
            self.command_name.append(command_name)
            self.commands.append(command)
            self.command_application.add_commands(command)
        return self

    def swap(self, command: "Command") -> None:
        """Swap an (existing) command with the given one."""
        command_name = command.config.name
        # if command with same name has been registered remove it
        if self.command_application.find(command_name):
            # no public API to do this yet
            del self.command_application.commands._commands[command_name]
        self.add(command)

    def run(self):
        """Run the cleo application."""
        return self.command_application.run()



from .Facade import Facade
from .Request import Request
from .Response import Response
from .Mail import Mail
from .Hash import Hash
from .Url import Url
from .Session import Session
from .View import View
from .Gate import Gate
from .Auth import Auth
from .Config import Config
from .Loader import Loader
from .Notification import Notification
from .Storage import Storage
from .Dump import Dump
from .Queue import Queue
from .Broadcast import Broadcast



class TableNotFound:
    def title(self):
        return "Table Not Found"

    def description(self):
        return "You are trying to make a query on a table that cannot be found. Check that :table migration exists and that migrations have been ran with 'python craft migrate' command."

    def regex(self):
        return r"no such table: (?P<table>(\w+))"


class MissingCSRFToken:
    def title(self):
        return "Missing CSRF Token"

    def description(self):
        return "You are trying to make a sensitive request without providing a CSRF token. Your request might be vulnerable to Cross Site Request Forgery. To resolve this issue you should use {{ csrf_field }} in HTML forms or add X-CSRF-TOKEN header in AJAX requests."

    def regex(self):
        return r"Missing CSRF Token"


class InvalidCSRFToken:
    def title(self):
        return "Invalid CSRF Token"

    def description(self):
        return "TODO"

    def regex(self):
        return r"Invalid CSRF Token"


class TemplateNotFound:
    def title(self):
        return "Template Not Found"

    def description(self):
        return """':template.html' view file has not been found in registered view locations. Please verify the spelling of the template and that it exists in locations declared in Kernel file. You can check
        available view locations with app.make('view.locations')."""

    def regex(self):
        return r"Template '(?P<template>(\w+))' not found"


class NoneResponse:
    def title(self):
        return "Response cannot be None"

    def description(self):
        return """Ensure that the controller method used in this request returned something. A controller method cannot return None or nothing.
        If you don't want to return a value you can return an empty string ''."""

    def regex(self):
        return r"Responses cannot be of type: None."


class RouteMiddlewareNotFound:
    def title(self):
        return "Did you register the middleware key in your Kernel.py file?"

    def description(self):
        return "Check your Kernel.py file inside your 'route_middleware' attribute and look for a :middleware key"

    def regex(self):
        return r"Could not find the \'(?P<middleware>(\w+))\' middleware key"



from tests import TestCase


class TestPackageProviderPublishing(TestCase):
    def test_publishing_migrations(self):
        self.craft(
            "package:publish", "test_package --r migrations --dry"
        ).assertOutputContains("1_create_some_table.py").assertOutputContains(
            "2_create_other_table.py"
        )



from typing import Any

class Storage:
    """File storage facade."""

    def add_driver(name: str, driver: str): ...
    def set_configuration(config: dict) -> "Storage": ...
    def get_driver(name: str = None) -> Any: ...
    def get_config_options(name: str = None) -> dict: ...
    def disk(name: str = "default") -> Any:
        """Get the file manager instance for the given disk name."""
        ...



from src.masonite.providers import Provider
from src.masonite.facades import Gate
from ..tasks.TaskTest import TaskTest


class AppProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        # Everyone (guests and authenticated users can view posts)
        Gate.define("view-posts", lambda user=None: True)
        Gate.define("display-admin", lambda user: user.email == "admin@gmail.com")

        self.application.make("scheduler").add(TaskTest().daily())

    def boot(self):
        pass



from tests import TestCase
from src.masonite.routes import Route
from src.masonite.controllers import Controller
from src.masonite.exceptions import RouteNotFoundException


class TestController(Controller):
    def simple(self):
        1 / 0

    def http(self):
        raise RouteNotFoundException()


class TestExceptionHandlerInDebug(TestCase):
    def setUp(self):
        super().setUp()
        self.handler = self.application.make("exception_handler")
        self.setRoutes(
            Route.get("/simple", TestController.simple),
            Route.get("/http", TestController.http),
        )
        # enable exceptions handling during all the tests of this class
        # because it is what is tested here
        self.withExceptionsHandling()

    def test_that_exception_event_is_fired(self):
        # with self.debugMode():
        #     self.get("/simple")
        pass

    def test_raising_simple_exception_renders_debug_error_page(self):
        with self.debugMode():
            self.get("/simple").assertError().assertContains(
                "ZeroDivisionError"
            ).assertContains("Exceptionite")

    def test_raising_http_exception_renders_debug_error_page(self):
        with self.debugMode():
            self.get("/http").assertError().assertContains(
                "RouteNotFoundException"
            ).assertContains("Exceptionite")

    def test_raising_exception_output_stack_trace_to_console(self):
        with self.debugMode():
            self.get("/simple")
            self.assertConsoleOutputContains(
                "ZeroDivisionError: division by zero"
            ).assertConsoleOutputContains("Stack Trace")

    def test_accepting_json_returns_debug_error_payload(self):
        with self.debugMode():
            self.withHeaders({"Accept": "application/json"}).get(
                "/simple"
            ).assertError().assertJsonPath("exception.type", "ZeroDivisionError")


class TestExceptionHandler(TestCase):
    """Test error handling in production mode, debug is False."""

    def setUp(self):
        super().setUp()
        self.handler = self.application.make("exception_handler")
        self.setRoutes(Route.get("/simple", TestController.simple))
        # enable exceptions handling during all the tests of this class
        # because it is what is tested here
        self.withExceptionsHandling()

    def test_that_exception_event_is_fired(self):
        pass

    def test_raising_simple_exception_renders_500_error_template(self):
        with self.debugMode(False):
            self.get("/simple").assertError().assertContains("Error 500")

    def test_raising_http_exception_renders_404_error_page(self):
        with self.debugMode(False):
            self.get("/http").assertNotFound().assertContains("Page Not Found")

    def test_raising_exception_does_not_output_stack_trace_to_console(self):
        with self.debugMode(False):
            self.get("/simple")
            self.assertConsoleEmpty()

    def test_accepting_json_returns_500_error_payload(self):
        with self.debugMode(False):
            self.withHeaders({"Accept": "application/json"}).get(
                "/simple"
            ).assertError().assertJson({"status": 500, "message": "division by zero"})



import pendulum
from src.masonite.tests import TestCase
from src.masonite.scheduling import Task


class MockTask(Task):
    run_every = "5 minutes"
    timezone = "America/New_York"


class TestScheduler(TestCase):
    def setUp(self):
        super().setUp()
        self.task = MockTask()

    def test_scheduler_should_run(self):
        assert self.task.run_every == "5 minutes"
        time = pendulum.now().on(2018, 5, 21).at(22, 5, 5)
        self.task._date = time
        assert self.task.should_run(time) == True

        time = pendulum.now().on(2018, 5, 21).at(22, 6, 5)
        self.task._date = time
        assert self.task.should_run(time) == False

    def test_scheduler_should_run_every_minute(self):
        self.task.run_every = "1 minute"
        time = pendulum.now().on(2018, 5, 21).at(22, 5, 5)
        self.task._date = time
        assert self.task.should_run(time) == True

        time = pendulum.now().on(2018, 5, 21).at(22, 6, 5)
        self.task._date = time
        assert self.task.should_run(time) == True

    def test_scheduler_should_run_every_2_minutes(self):
        self.task.run_every = "2 minutes"
        time = pendulum.now().on(2018, 5, 21).at(14, 56, 5)
        self.task._date = time
        assert self.task.should_run(time) == True

        time = pendulum.now().on(2018, 5, 21).at(14, 58, 5)
        self.task._date = time
        assert self.task.should_run(time) == True

    def test_scheduler_should_run_every_hour(self):
        self.task.run_every = "1 hour"
        time = pendulum.now().on(2018, 5, 21).at(2, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == True

        time = pendulum.now().on(2018, 5, 21).at(3, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == True

        self.task.run_every = "2 hours"
        time = pendulum.now().on(2018, 5, 21).at(2, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == True

        self.task.run_every = "2 hours"
        time = pendulum.now().on(2018, 5, 21).at(3, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == False

        time = pendulum.now().on(2018, 5, 21).at(4, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == True

    def test_scheduler_should_run_every_days(self):
        self.task.run_every = "2 days"
        time = pendulum.now().on(2018, 5, 21).at(0, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == False

        time = pendulum.now().on(2018, 5, 23).at(0, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == False

        self.task.run_at = "5:30"
        time = pendulum.now().on(2018, 5, 22).at(5, 30, 0)
        self.task._date = time
        assert self.task.should_run(time) == True

        self.task.run_at = "5:35"
        time = pendulum.now().on(2018, 5, 22).at(5, 30, 0)
        self.task._date = time
        assert self.task.should_run(time) == False

    def test_scheduler_should_run_every_months(self):
        self.task.run_every = "2 months"
        time = pendulum.now().on(2018, 1, 1).at(0, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == False

        time = pendulum.now().on(2018, 2, 1).at(0, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == True

        time = pendulum.now().on(2018, 2, 1).at(10, 0, 1)
        self.task._date = time
        assert self.task.should_run(time) == False

        self.task.run_at = "5:30"
        time = pendulum.now().on(2018, 2, 1).at(5, 30, 0)
        self.task._date = time
        assert self.task.should_run(time) == False

    def test_twice_daily_at_correct_time(self):
        time = pendulum.now().on(2018, 1, 1).at(1, 20, 5)
        self.task.run_every = ""
        self.task.twice_daily = (1, 13)
        self.task._date = time

        assert self.task.should_run()

        time = pendulum.now().on(2018, 1, 1).at(13, 20, 5)
        self.task._date = time
        assert self.task.should_run()

    def test_twice_daily_at_incorrect_time(self):
        time = pendulum.now().on(2018, 1, 1).at(12, 20, 5)
        self.task.run_every = ""
        self.task.twice_daily = (1, 13)
        self.task._date = time

        assert self.task.should_run() is False

    def test_run_at(self):
        self.task.run_every = ""
        self.task.run_at = "13:00"

        time = pendulum.now().on(2018, 1, 1).at(13, 0, 5)
        self.task._date = time

        self.task.run_at = "13:05"

        time = pendulum.now().on(2018, 1, 1).at(13, 5, 5)
        self.task._date = time

        assert self.task.should_run() is True

        time = pendulum.now().on(2018, 1, 1).at(13, 6, 5)
        self.task._date = time

        assert self.task.should_run() is False

    def test_method_calls(self):
        task = MockTask()
        task.at("13:00")

        time = pendulum.now().on(2018, 1, 1).at(13, 0, 5)
        task._date = time

        assert task.should_run(time) == True

        task = MockTask()
        task.every_minute()

        time = pendulum.now().on(2018, 5, 21).at(22, 5, 5)
        task._date = time
        assert task.should_run(time) == True

    def test_should_run_task_immediately_by_class(self):
        self.craft("schedule:run", "--task TaskTest --force").assertSuccess()

    def test_should_run_task_immediately_by_name(self):
        self.craft("schedule:run", "--task task_test --force").assertSuccess()



"""Notifications Settings."""
from src.masonite.environment import env


DRIVERS = {
    "slack": {
        "token": env("SLACK_TOKEN", ""),  # used for API mode
        "webhook": env("SLACK_WEBHOOK", ""),  # used for webhook mode
    },
    "vonage": {
        "key": env("VONAGE_KEY", ""),
        "secret": env("VONAGE_SECRET", ""),
        "sms_from": env("VONAGE_SMS_FROM", "+33000000000"),
    },
    "database": {
        "connection": "sqlite",
        "table": "notifications",
    },
}

DRY = False



class PresetsCapsule:
    def __init__(self):
        self.presets = {}

    def add(self, preset):
        self.presets.update({preset.key: preset})
        return self

    def get_presets(self):
        return self.presets.values()

    def get_presets_keys(self):
        return self.presets.keys()

    def run(self, key):
        return self.presets.get(key).install()



from tests import TestCase
from src.masonite.mail import Mailable


class Welcome(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )


class Other(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Other")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
            .driver("terminal")
        )


class TestTerminalDriver(TestCase):
    def test_send_mailable(self):
        self.application.make("mail").mailable(
            Welcome().attach("invoice", "tests/integrations/storage/invoice.pdf")
        ).send(driver="terminal")
        # TODO: once PR #560 merged add
        # self.assertConsoleOutputContains("From: joe@masoniteproject.com")
        # self.assertConsoleOutputContains("To: idmann509@gmail.com")
        # self.assertConsoleOutputContains("Subject: Masonite 4")
        # self.assertConsoleOutputContains("Attachment 0: invoice")

    def test_define_driver_with_mailable(self):
        self.application.make("mail").mailable(
            Other().attach("invoice", "tests/integrations/storage/invoice.pdf")
        ).send()
        # TODO: once PR #560 merged add
        # self.assertConsoleOutputContains("From: no-reply@masonite.com")
        # self.assertConsoleOutputContains("To: idmann509@gmail.com")
        # self.assertConsoleOutputContains("Subject: Other")
        # self.assertConsoleOutputContains("Attachment 0: invoice")



"""New Task Command """
import os
import inflection

from ...utils.filesystem import make_directory, get_module_dir, render_stub_file
from ...utils.location import base_path
from ...utils.str import as_filepath
from ...commands.Command import Command


class MakeTaskCommand(Command):
    """
    Create a new task
    task
        {name : Name of the task you want to create}
        {--d|--directory=? : Override the directory to create the task in}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        output = render_stub_file(self.get_stub_task_path(), name)

        relative_file_name = os.path.join(
            self.option("directory") or as_filepath(self.app.make("tasks.location")),
            f"{name}.py",
        )
        filepath = base_path(relative_file_name)

        if os.path.exists(relative_file_name) and not self.option("force"):
            self.warning(
                f"{relative_file_name} already exists! Run the command with -f (force) to override."
            )
            return -1

        make_directory(filepath)
        with open(filepath, "w") as fp:
            fp.write(output)

        self.info(f"Task Created ({relative_file_name})")

    def get_stub_task_path(self):
        return os.path.join(get_module_dir(__file__), "../../stubs/scheduling/Task.py")



from tests import TestCase


class TestAuthentication(TestCase):
    def setUp(self):
        super().setUp()
        self.auth = self.application.make("auth")
        self.make_request()
        self.make_response()

    def test_attempt(self):
        self.assertTrue(self.auth.attempt("idmann509@gmail.com", "secret"))
        self.assertFalse(self.auth.attempt("idmann509@gmail.com", "secret1"))

    def test_auth_class_registers_cookie(self):
        self.auth.guard("web").attempt("idmann509@gmail.com", "secret")
        self.assertTrue(self.application.make("response").cookie("token"))

    def test_logout(self):
        self.application.make("auth").guard("web").attempt(
            "idmann509@gmail.com", "secret"
        )

        self.assertTrue(self.application.make("response").cookie("token"))

        self.application.make("auth").guard("web").logout()
        self.assertTrue(
            self.application.make("response").cookie_jar.deleted_cookies["token"]
        )

    def test_attempt_by_id(self):
        self.application.make("auth").guard("web").attempt_by_id(1)

        self.assertTrue(self.application.make("response").cookie("token"))

        self.application.make("auth").guard("web").logout()
        self.assertFalse(self.application.make("response").cookie("token"))

    def test_attempt_by_id_once(self):
        self.application.make("auth").guard("web").attempt_by_id(1, once=True)

        self.assertIsNone(self.application.make("request").cookie("token"))



from typing import Any

class Hash:
    def add_driver(name: str, driver: Any): ...
    def set_configuration(config: dict) -> "Hash": ...
    def get_driver(name: str = None) -> Any: ...
    def get_config_options(driver: str = None) -> dict: ...
    def make(string: str, options: dict = {}, driver: str = None) -> str:
        """Hash a string and return as string based on configured hashing protocol."""
        ...
    def make_bytes(string: str, options: dict = {}, driver: str = None) -> bytes:
        """Hash a string and return as bytes based on configured hashing protocol."""
        ...
    def check(
        self,
        plain_string: str,
        hashed_string: str,
        options: dict = {},
        driver: str = None,
    ) -> bool:
        """Verify that a given string matches its hashed version (based on configured hashing protocol)."""
        ...
    def needs_rehash(
        hashed_string: str, options: dict = {}, driver: str = None
    ) -> bool:
        """Verify that a given hash needs to be hashed again because parameters for generating
        the hash have changed."""
        ...



from .RuleEnclosure import RuleEnclosure
from .MessageBag import MessageBag
from .Validator import (
    BaseValidation,
    ValidationFactory,
    Validator,
    accepted,
    active_domain,
    after_today,
    before_today,
    confirmed,
    contains,
    date,
    different,
    distinct,
    does_not,
    email,
    equals,
    exists,
    exists_in_db,
    file,
    greater_than,
    image,
    in_range,
    ip,
    is_future,
    is_list,
    is_in,
    is_past,
    isnt,
    json,
    length,
    less_than,
    none,
    numeric,
    phone,
    postal_code,
    regex,
    required,
    required_if,
    required_with,
    string,
    strong,
    timezone,
    truthy,
    unique_in_db,
    uuid,
    video,
    when,
)



"""Cache Config"""

import os

BROADCASTS = {
    "default": "pusher",
    "pusher": {
        "driver": "pusher",
        "client": os.getenv("PUSHER_CLIENT"),
        "app_id": os.getenv("PUSHER_APP_ID"),
        "secret": os.getenv("PUSHER_SECRET"),
        "cluster": os.getenv("PUSHER_CLUSTER"),
        "ssl": False,
    },
}



from tests import TestCase


class TestResponseDownload(TestCase):
    def setUp(self):
        super().setUp()
        self.response = self.make_response()

    def test_download(self):
        self.response.download(
            "invoice", "tests/integrations/storage/invoice.pdf", force=True
        )
        self.assertTrue(self.response.header("Content-Disposition"))
        self.assertEqual(
            self.response.header("Content-Disposition"),
            'attachment; filename="invoice.pdf"',
        )



from inflection import titleize

from .Header import Header


class HeaderBag:
    def __init__(self):
        self.bag: dict = {}

    def add(self, header: "Header") -> None:
        self.bag.update({self.convert_name(header.name): header})

    def add_if_not_exists(self, header: "Header") -> None:
        if self.convert_name(header.name) in self.bag:
            return None

        self.bag.update({self.convert_name(header.name): header})

    def get(self, name: str) -> "Header":
        return self.bag.get(self.convert_name(name), "")

    def render(self) -> list:
        response = []
        for name, header in self.bag.items():
            response.append((name, header.value))

        return response

    def __contains__(self, name: str) -> bool:
        return self.convert_name(name) in self.bag.keys()

    def convert_name(self, name: str) -> str:
        return titleize(name.replace("HTTP_", "")).replace(" ", "-")

    def load(self, environ: dict) -> None:
        for key, value in environ.items():
            if key.startswith("HTTP_"):
                self.add(Header(key, value))

    def to_dict(self) -> dict:
        dic = {}
        for name, header in self.bag.items():
            dic.update({name: header.value})

        return dic

    def __getitem__(self, key: str) -> "Header":
        return self.bag[self.convert_name(key)]



from tests import TestCase
from src.masonite.mail import Mailable


class Welcome(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("text from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )


class Other(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .text("text from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )


class TestSMTPDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.fake("mail")

    def tearDown(self):
        super().tearDown()
        self.restore("mail")

    def test_mock_mail(self):
        welcome_email = self.application.make("mail").mailable(Welcome()).send()
        (
            welcome_email.seeEmailCc("")
            .seeEmailBcc("")
            .seeEmailContains("Hello from Masonite!")
            .seeEmailContains("text from Masonite!")
            .seeEmailFrom("joe@masoniteproject.com")
            .seeEmailCountEquals(1)
            .send()
            .seeEmailCountEquals(2)
        )

    def test_mock_mail_sending(self):
        welcome_email = self.application.make("mail").mailable(Welcome())
        (welcome_email.seeEmailWasNotSent().send().seeEmailWasSent())

    def test_mock_can_also_handle_options(self):
        other_email = self.application.make("mail").mailable(Other())
        other_email.send().seeEmailFrom("no-reply@masonite.com")



from masonite.notification import Notification
from masonite.mail import Mailable


class __class__(Notification, Mailable):
    def to_mail(self, notifiable):
        return (
            self.to(notifiable.email)
            .subject("Masonite 4")
            .from_("hello@email.com")
            .text(f"Hello {notifiable.name}")
        )

    def via(self, notifiable):
        return ["mail"]



import pendulum
import os
import pytest
import sys

from tests import TestCase
from tests.integrations.controllers.WelcomeController import WelcomeController
from masoniteorm.models import Model
from src.masonite.routes import Route
from src.masonite.authentication import Authenticates
from src.masonite.middleware import EncryptCookies


class User(Model, Authenticates):
    pass


class CustomTestResponse:
    def assertCustom(self):
        assert 1
        return self


class OtherCustomTestResponse:
    def assertOtherCustom(self):
        assert 2
        return self


class TestTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            Route.get("/", "WelcomeController@show").name("home"),
        )

    def tearDown(self):
        super().tearDown()
        self.restoreTime()

    def test_add_routes(self):
        self.assertEqual(len(self.application.make("router").routes), 1)
        self.addRoutes(Route.get("/some-route", "WelcomeController@show"))
        self.assertEqual(len(self.application.make("router").routes), 2)

    def test_use_custom_test_response(self):
        self.application.make("tests.response").add(
            CustomTestResponse, OtherCustomTestResponse
        )
        # can use default assertions and custom from different classes
        self.get("/").assertContains("Welcome").assertCustom().assertOtherCustom()

    def test_fake_time(self):
        given_date = pendulum.datetime(2015, 2, 5)
        self.fakeTime(given_date)
        self.assertEqual(pendulum.now(), given_date)
        self.restoreTime()
        self.assertNotEqual(pendulum.now(), given_date)

    def test_fake_time_tomorrow(self):
        tomorrow = pendulum.tomorrow()
        self.fakeTimeTomorrow()
        self.assertEqual(pendulum.now(), tomorrow)

    def test_fake_time_yesterday(self):
        yesterday = pendulum.yesterday()
        self.fakeTimeYesterday()
        self.assertEqual(pendulum.now(), yesterday)

    def test_fake_time_in_future(self):
        real_now = pendulum.now()
        self.fakeTimeInFuture(10)
        self.assertEqual(pendulum.now().diff(real_now).in_days(), 10)
        self.assertGreater(pendulum.now(), real_now)

        self.fakeTimeInFuture(1, "months")
        self.assertEqual(pendulum.now().diff(real_now).in_months(), 1)

    # def test_fake_time_in_past(self):
    #     real_now = pendulum.now()
    #     self.fakeTimeInPast(10)
    #     self.assertEqual(pendulum.now().diff(real_now).in_days(), 10)
    #     self.assertLess(pendulum.now(), real_now)

    #     self.fakeTimeInPast(3, "hours")
    #     self.assertEqual(real_now.hour - pendulum.now().hour, 3)

    def test_env_context_manager(self):
        initial = os.getenv("APP_ENV")
        with self.env("custom_env_for_this_test"):
            assert os.getenv("APP_ENV") == "custom_env_for_this_test"
        assert os.getenv("APP_ENV") == initial

    def test_debugMode_context_manager(self):
        with self.debugMode():
            self.assertTrue(self.application.is_debug())
        with self.debugMode(False):
            self.assertFalse(self.application.is_debug())

    # Skipped to avoid polluting test outputs, but kept there
    # def test_dump(self):
    #     self.dump("test", "dump test")

    # def test_dump_session(self):
    #     self.get("/").dumpSession()

    # def test_dump_headers(self):
    #     self.get("/").dumpRequestHeaders().dumpResponseHeaders()


class TestTestingAssertions(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            Route.get("/", "WelcomeController@show").name("home"),
            Route.get("/test", "WelcomeController@show").name("test"),
            Route.get("/view", "WelcomeController@view").name("view"),
            Route.get("/view-context", "WelcomeController@view_with_context").name(
                "view_with_context"
            ),
            Route.get("/test-404", "WelcomeController@not_found").name("not_found"),
            Route.get("/test-creation", "WelcomeController@create").name("create"),
            Route.get("/test-unauthorized", "WelcomeController@unauthorized").name(
                "unauthorized"
            ),
            Route.get("/test-forbidden", "WelcomeController@forbidden").name(
                "forbidden"
            ),
            Route.get("/test-empty", "WelcomeController@empty").name("empty"),
            Route.get(
                "/test-response-header", "WelcomeController@response_with_headers"
            ),
            Route.get("/test-redirect-1", "WelcomeController@redirect_url"),
            Route.get("/test-redirect-2", "WelcomeController@redirect_route"),
            Route.get("/test-redirect-3", "WelcomeController@redirect_route_params"),
            Route.get("/test/@id", "WelcomeController@with_params").name("test_params"),
            Route.get("/test-json", "WelcomeController@json").name("json"),
            Route.get("/test-session", "WelcomeController@session").name("session"),
            Route.get(
                "/test-session-errors", "WelcomeController@session_with_errors"
            ).name("session"),
            Route.get("/test-session-2", "WelcomeController@session2").name("session2"),
            Route.get("/test-authenticates", "WelcomeController@auth").name("auth"),
        )

    def test_assert_contains(self):
        self.get("/").assertContains("Welcome")
        self.get("/").assertNotContains("hello")

    def test_assert_is_named(self):
        self.get("/test").assertIsNamed("test")
        self.get("/test").assertIsNotNamed("welcome")

    def test_assert_not_found(self):
        self.get("/test-404").assertNotFound()

    def test_assert_is_status(self):
        self.get("/test").assertIsStatus(200)

    def test_assert_ok(self):
        self.get("/test").assertOk()

    def test_assert_created(self):
        self.get("/test-creation").assertCreated()

    def test_assert_unauthorized(self):
        self.get("/test-unauthorized").assertUnauthorized()

    def test_assert_forbidden(self):
        self.get("/test-forbidden").assertForbidden()

    def test_assert_no_content(self):
        self.get("/test-empty").assertNoContent()

    def test_assert_cookie(self):
        self.withCookies({"test": "value"}).get("/").assertCookie("test")

    def test_assert_cookie_value(self):
        self.withCookies({"test": "value"}).get("/").assertCookie("test", "value")

    def test_assert_cookie_missing(self):
        self.get("/").assertCookieMissing("test")

    def test_assert_plain_cookie(self):
        self.withCookies({"test": "value"}).get("/").assertPlainCookie("test")

    def test_assert_has_header(self):
        self.get("/test-response-header").assertHasHeader("TEST")
        self.get("/test-response-header").assertHasHeader("TEST", "value")

    def test_assert_header_missing(self):
        self.get("/").assertHeaderMissing("X-Test")

    def test_with_headers(self):
        request = (
            self.withHeaders({"X-Test": "value", "HTTP_CUSTOM_1": 0}).get("/").request
        )
        assert request.header("X-Test") == "value"
        assert request.header("Custom-1") == 0

    def test_assert_redirect_to_url(self):
        self.get("/test-redirect-1").assertRedirect("/")

    def test_assert_redirect_to_route(self):
        self.get("/test-redirect-2").assertRedirect(name="test")
        self.get("/test-redirect-3").assertRedirect(
            name="test_params", params={"id": 1}
        )

    def test_assert_session_has(self):
        self.get("/test-session").assertSessionHas("key")
        self.get("/test-session").assertSessionHas("key", "value")

    def test_assert_session_has_errors(self):
        self.get("/test-session-errors").assertSessionHasErrors()
        self.get("/test-session-errors").assertSessionHasErrors(["email"])
        self.get("/test-session-errors").assertSessionHasErrors(["email", "password"])

    def test_assert_session_has_no_errors(self):
        self.get("/test-session").assertSessionHasNoErrors()
        self.get("/test-session-errors").assertSessionHasNoErrors(["name"])

    def test_assert_session_missing(self):
        self.get("/").assertSessionMissing("some_test_key")

    # def test_with_session(self):
    #     self.withSession({"key1": "value1"}).get("/").assertSessionHas("key1", "value1")

    def test_assert_view_is(self):
        self.get("/view").assertViewIs("welcome")

    def test_assert_view_has(self):
        self.get("/view-context").assertViewHas("count")
        self.get("/view-context").assertViewHas("count", 1)
        self.get("/view-context").assertViewHas("users", ["John", "Joe"])
        self.get("/view-context").assertViewHas("other_key.nested")
        self.get("/view-context").assertViewHas("other_key.nested", 1)

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHas("not_in_view")
        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHas("not_in_view", 3)

    def test_assert_view_helpers_raise_error_if_not_rendering_a_view(self):
        # json response
        with self.assertRaises(ValueError):
            self.get("/test-json").assertViewIs("test")
        # string response
        self.get("/test").assertViewIs("welcome")

    def test_assert_view_has_exact(self):
        self.get("/view-context").assertViewHasExact(["users", "count", "other_key"])
        self.get("/view-context").assertViewHasExact(
            {"count": 1, "users": ["John", "Joe"], "other_key": {"nested": 1}}
        )

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHasExact(
                ["users", "count", "not in data"]
            )

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHasExact({"count": 1})

    def test_assert_view_missing(self):
        self.get("/view-context").assertViewMissing("not in data")

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewMissing("users")

    def test_assert_guest(self):
        self.get("/test").assertGuest()

    def test_assert_authenticated(self):
        self.get("/test-authenticates").assertAuthenticated()
        user = User.find(1)
        self.get("/test-authenticates").assertAuthenticated(user)

    def test_acting_as(self):
        user = User.find(1)
        self.actingAs(user).get("/test").assertAuthenticated(user)

    def test_assert_has_controller(self):
        self.get("/test").assertHasController("WelcomeController@show")
        self.get("/test").assertHasController(WelcomeController)

    def test_assert_route_has_parameter(self):
        self.get("/test/3").assertRouteHasParameter("id")
        with self.assertRaises(AssertionError):
            self.get("/test/3").assertRouteHasParameter("key")
        self.get("/test/3").assertRouteHasParameter("id", 3)
        with self.assertRaises(AssertionError):
            self.get("/test/3").assertRouteHasParameter("id", 4)

    def test_assert_has_route_middleware(self):
        self.get("/test").assertHasRouteMiddleware("web")

    def test_assert_has_http_middleware(self):
        self.get("/test").assertHasHttpMiddleware(EncryptCookies)

    def test_assert_json(self):
        self.get("/test-json").assertJson({"key": "value"})
        # works also in a nested path
        self.get("/test-json").assertJson(
            {"other_key": {"nested": 1, "nested_again": {"a": 1, "b": 2}}}
        )

    def test_json_assertions_fail_when_response_not_json(self):
        with self.assertRaises(ValueError):
            self.get("/view").assertJson({"key": "value"})

    def test_assert_json_path(self):
        self.get("/test-json").assertJsonPath("key2", [1, 2])
        self.get("/test-json").assertJsonPath("other_key.nested", 1)
        self.get("/test-json").assertJsonPath("other_key.nested_again.b", 2)
        self.get("/test-json").assertJsonPath(
            "other_key.nested_again", {"a": 1, "b": 2}
        )

    def test_assert_json_count(self):
        self.get("/test-json").assertJsonCount(3)
        self.get("/test-json").assertJsonCount(2, key="other_key")

    def test_assert_json_exact(self):
        self.get("/test-json").assertJsonExact(
            {
                "key": "value",
                "key2": [1, 2],
                "other_key": {
                    "nested": 1,
                    "nested_again": {"a": 1, "b": 2},
                },
            }
        )

    def test_assert_json_missing(self):
        self.get("/test-json").assertJsonMissing("key3")
        self.get("/test-json").assertJsonMissing("some_key.nested")
        with self.assertRaises(AssertionError):
            self.get("/test-json").assertJsonMissing("other_key.nested")

    def test_assert_database_count(self):
        self.assertDatabaseCount("users", 1)

    def test_assert_database_has(self):
        self.assertDatabaseHas("users", {"name": "Joe"})

    def test_assert_database_missing(self):
        self.assertDatabaseMissing(
            "users", {"name": "John", "email": "john@example.com"}
        )

    def test_assert_console_empty(self):
        self.assertConsoleEmpty()

    def test_assert_console_not_empty(self):
        print("test")
        self.assertConsoleNotEmpty()

    def test_assert_console_exact_output(self):
        print("Hello World !")
        self.assertConsoleExactOutput("Hello World !\n")

    def test_assert_console_output_missing(self):
        print("Hello World !")
        self.assertConsoleOutputMissing("Not there")

    def test_assert_console_output_contains(self):
        print("Hello World !")
        self.assertConsoleOutputContains("Hello")

    def test_assert_console_has_errors(self):
        print("Fatal Error !", file=sys.stderr)
        self.assertConsoleHasErrors()

    def test_assert_console_exact_error(self):
        print("Fatal Error !", file=sys.stderr)
        self.assertConsoleExactError("Fatal Error !\n")

    def test_assert_console_error_contains(self):
        print("Fatal Error !", file=sys.stderr)
        self.assertConsoleErrorContains("Error")

    def test_make_multiple_output_assertions(self):
        print("Hello World !")
        self.assertConsoleExactOutput("Hello World !\n")
        self.assertConsoleOutputContains("Hello")



from ...facades import Auth
from .. import Middleware


class GuardMiddleware(Middleware):
    def before(self, request, _, guard):
        Auth.guard(guard)
        return request

    def after(self, request, _, __):
        return request



from .HTTPRoute import HTTPRoute
from ..utils.collections import flatten
from ..utils.str import modularize
from ..controllers import RedirectController


class Route:

    routes = []
    compilers = {
        "int": r"(\d+)",
        "integer": r"(\d+)",
        "string": r"([a-zA-Z]+)",
        "default": r"([\w.-]+)",
        "signed": r"([\w\-=]+)",
    }
    controllers_locations = []

    def __init__(self):
        pass

    @classmethod
    def get(self, url, controller, module_location=None, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["get"],
            compilers=self.compilers,
            controllers_locations=module_location or self.controllers_locations,
            **options,
        )

    @classmethod
    def post(self, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["post"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def put(self, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["put"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def patch(self, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["patch"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def delete(self, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["delete"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def options(self, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=["options"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def default(self, url, controller, **options):
        return self

    @classmethod
    def redirect(self, url, new_url, **options):
        return HTTPRoute(
            url,
            RedirectController.redirect,
            request_method=["get"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            controller_bindings=[new_url, options.get("status", 302)],
            **options,
        )

    @classmethod
    def permanent_redirect(self, url, new_url, **options):
        return HTTPRoute(
            url,
            RedirectController.redirect,
            request_method=["get"],
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            controller_bindings=[new_url, 301],
            **options,
        )

    @classmethod
    def match(self, request_methods, url, controller, **options):
        return HTTPRoute(
            url,
            controller,
            request_method=request_methods,
            compilers=self.compilers,
            controllers_locations=self.controllers_locations,
            **options,
        )

    @classmethod
    def group(self, *routes, **options):
        inner = []
        for route in flatten(routes):
            if options.get("prefix"):
                if route.url == "" or route.url == "/":
                    route.url = options.get("prefix")
                else:
                    route.url = options.get("prefix") + route.url

                route.compile_route_to_regex()

            if options.get("name"):
                route._name = options.get("name") + route._name

            if options.get("domain"):
                route.domain(options.get("domain"))

            if options.get("middleware"):
                middleware = route.list_middleware
                middleware = options.get("middleware", []) + middleware

                route.set_middleware(middleware)

            inner.append(route)
        self.routes = inner
        return inner

    @classmethod
    def resource(self, base_url, controller):
        return [
            self.get(f"/{base_url}", f"{controller}@index").name(f"{base_url}.index"),
            self.get(f"/{base_url}/create", f"{controller}@create").name(
                f"{base_url}.create"
            ),
            self.post(f"/{base_url}", f"{controller}@store").name(f"{base_url}.store"),
            self.get(f"/{base_url}/@id", f"{controller}@show").name(f"{base_url}.show"),
            self.get(f"/{base_url}/@id/edit", f"{controller}@edit").name(
                f"{base_url}.edit"
            ),
            self.match(
                ["put", "patch"], f"/{base_url}/@id", f"{controller}@update"
            ).name(f"{base_url}.update"),
            self.delete(f"/{base_url}/@id", f"{controller}@destroy").name(
                f"{base_url}.destroy"
            ),
        ]

    @classmethod
    def api(self, base_url, controller):
        return [
            self.get(f"/{base_url}", f"{controller}@index").name(f"{base_url}.index"),
            self.post(f"/{base_url}", f"{controller}@store").name(f"{base_url}.store"),
            self.get(f"/{base_url}/@id", f"{controller}@show").name(f"{base_url}.show"),
            self.match(
                ["put", "patch"], f"/{base_url}/@id", f"{controller}@update"
            ).name(f"{base_url}.update"),
            self.delete(f"/{base_url}/@id", f"{controller}@destroy").name(
                f"{base_url}.destroy"
            ),
        ]

    @classmethod
    def compile(self, key, to=""):
        self.compilers.update({key: to})
        return self

    @classmethod
    def set_controller_locations(self, *controllers_locations):
        self.controllers_locations = list(map(modularize, controllers_locations))
        return self

    @classmethod
    def add_controller_locations(self, *controllers_locations):
        self.controllers_locations.extend(list(map(modularize, controllers_locations)))
        return self



from typing import Any

class Queue:
    def add_driver(name: str, driver: str): ...
    def get_driver(name: str) -> Any: ...
    def set_configuration(config: dict) -> "Queue": ...
    def get_config_options(driver: str) -> dict: ...
    def push(*jobs, **options) -> None: ...
    def consume(options: dict) -> Any: ...
    def retry(options: dict) -> Any: ...



from masoniteorm.migrations import Migration


class CreateUsersTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("users") as table:
            table.increments("id")
            table.string("name")
            table.string("email").unique()
            table.string("password")
            table.string("second_password").nullable()
            table.string("remember_token").nullable()
            table.string("phone").nullable()
            table.timestamp("verified_at").nullable()
            table.timestamps()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("users")



from tests import TestCase

from src.masonite.notification import Notification, Notifiable, SlackMessage
from src.masonite.mail import Mailable
from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
        )

    def via(self, notifiable):
        return ["mail"]


class OtherNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Other")
            .from_("sam@masoniteproject.com")
            .text("Hello again!")
        )

    def via(self, notifiable):
        return ["mail"]


class OrderNotification(Notification):
    def __init__(self, order_id):
        self.order_id = order_id

    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject(f"Order {self.order_id} shipped !")
            .from_("sam@masoniteproject.com")
            .text(f"{notifiable.name}, your order has been shipped")
        )

    def to_slack(self, notifiable):
        return SlackMessage().text(f"Order {self.order_id} has been shipped !")

    def via(self, notifiable):
        return ["mail", "slack"]


class TestMockNotification(TestCase):
    def setUp(self):
        super().setUp()
        self.fake("notification")

    def tearDown(self):
        super().tearDown()
        self.restore("notification")

    def test_assert_nothing_sent(self):
        notification = self.application.make("notification")
        notification.assertNothingSent()

    def test_assert_count(self):
        notification = self.application.make("notification")
        notification.assertCount(0)
        notification.route("mail", "test@mail.com").send(WelcomeNotification())
        notification.assertCount(1)
        notification.route("mail", "test2@mail.com").send(WelcomeNotification())
        notification.assertCount(2)

    def test_reset_count(self):
        notification = self.application.make("notification")
        notification.assertNothingSent()
        notification.route("mail", "test@mail.com").send(WelcomeNotification())
        notification.resetCount()
        notification.assertNothingSent()

    def test_assert_sent_to_with_anonymous(self):
        notification = self.application.make("notification")
        notification.route("mail", "test@mail.com").send(WelcomeNotification())
        notification.assertSentTo("test@mail.com", WelcomeNotification)

        notification.route("vonage", "123456").route("slack", "#general").send(
            WelcomeNotification()
        )
        notification.assertSentTo("123456", WelcomeNotification)
        notification.assertSentTo("#general", WelcomeNotification)

    def test_assert_not_sent_to(self):
        notification = self.application.make("notification")
        notification.resetCount()
        notification.assertNotSentTo("test@mail.com", WelcomeNotification)
        notification.route("vonage", "123456").send(OtherNotification())
        notification.assertNotSentTo("123456", WelcomeNotification)
        notification.assertNotSentTo("test@mail.com", OtherNotification)

    def test_assert_sent_to_with_notifiable(self):
        notification = self.application.make("notification")
        user = User.find(1)
        user.notify(WelcomeNotification())
        notification.assertSentTo(user, WelcomeNotification)
        user.notify(OtherNotification())
        notification.assertSentTo(user, OtherNotification)
        notification.assertCount(2)

    def test_assert_sent_to_with_count(self):
        notification = self.application.make("notification")
        user = User.find(1)
        user.notify(WelcomeNotification())
        user.notify(WelcomeNotification())
        notification.assertSentTo(user, WelcomeNotification, count=2)

        user.notify(OtherNotification())
        user.notify(OtherNotification())
        with self.assertRaises(AssertionError):
            notification.assertSentTo(user, OtherNotification, count=1)

    def test_assert_with_assertions_on_notification(self):
        user = User.find(1)
        user.notify(OrderNotification(6))
        self.application.make("notification").assertSentTo(
            user,
            OrderNotification,
            lambda user, notification: (
                notification.assertSentVia("mail", "slack")
                .assertEqual(notification.order_id, 6)
                .assertEqual(
                    notification.to_mail(user).get_options().get("subject"),
                    "Order 6 shipped !",
                )
                .assertIn(
                    user.name,
                    notification.to_mail(user).get_options().get("text_content"),
                )
            ),
        )

    def test_last_notification(self):
        notification = self.application.make("notification")
        message = WelcomeNotification()
        notification.route("mail", "test@mail.com").send(message)
        assert message == notification.last()

    def test_assert_last(self):
        self.application.make("notification").route("mail", "test@mail.com").route(
            "slack", "#general"
        ).send(OrderNotification(10))
        self.application.make("notification").assertLast(
            lambda user, notif: (
                notif.assertSentVia("mail")
                .assertEqual(notif.order_id, 10)
                .assertEqual(
                    notif.to_slack(user).get_options().get("text"),
                    "Order 10 has been shipped !",
                )
            )
        )



class DriverNotFound(Exception):
    pass


class DriverLibraryNotFound(Exception):
    pass


class FileTypeException(Exception):
    pass


class RequiredContainerBindingNotFound(Exception):
    pass


class MissingContainerBindingNotFound(Exception):
    pass


class UnacceptableDriverType(Exception):
    pass


class ContainerError(Exception):
    pass


class InvalidCSRFToken(Exception):
    pass


class InvalidHTTPStatusCode(Exception):
    pass


class RouteMiddlewareNotFound(Exception):
    pass


class ResponseError(Exception):
    pass


class InvalidAutoloadPath(Exception):
    pass


class AutoloadContainerOverwrite(Exception):
    pass


class InvalidSecretKey(Exception):
    pass


class InvalidToken(Exception):
    pass


class StrictContainerException(Exception):
    pass


class InvalidRouteCompileException(Exception):
    pass


class RouteException(Exception):
    pass


class RouteNotFoundException(Exception):
    is_http_exception = True

    def __init__(self, message):
        super().__init__(message)
        self.message = message or "Route Not Found"

    def get_response(self):
        return self.message

    def get_status(self):
        return 404


class DebugException(Exception):
    pass


class DumpException(Exception):
    pass


class ViewException(Exception):
    pass


class QueueException(Exception):
    pass


class AmbiguousError(Exception):
    pass


class ProjectLimitReached(Exception):
    pass


class ProjectProviderTimeout(Exception):
    pass


class ProjectProviderHttpError(Exception):
    pass


class ProjectTargetNotEmpty(Exception):
    pass


class NotificationException(Exception):
    pass


class ModelNotFoundException(Exception):
    is_http_exception = True

    def get_response(self):
        return "Model Not Found"

    def get_status(self):
        return 404


class AuthorizationException(Exception):
    is_http_exception = True

    def __init__(self, message="", status=403):
        super().__init__(message)
        self.message = message or "Action not authorized"
        self.status = status or 403

    def get_response(self):
        return self.message

    def get_status(self):
        return self.status


class GateDoesNotExist(Exception):
    pass


class PolicyDoesNotExist(Exception):
    pass


class MixManifestNotFound(Exception):
    pass


class MixFileNotFound(Exception):
    pass


class InvalidConfigurationLocation(Exception):
    pass


class InvalidConfigurationSetup(Exception):
    pass


class InvalidPackageName(Exception):
    pass


class LoaderNotFound(Exception):
    pass



from masoniteorm.migrations import Migration


class CreateQueueJobsTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("jobs") as table:
            table.increments("id")
            table.string("name")
            table.string("queue")
            table.binary("payload")
            table.integer("attempts")
            table.timestamp("ran_at").nullable()
            table.timestamp("available_at", now=True).nullable()
            table.timestamp("reserved_at").nullable()
            table.timestamp("created_at", now=True).nullable()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("jobs")



from tests import TestCase
from src.masonite.facades import Mail, View


class TestFacades(TestCase):
    def test_mail_facade(self):
        self.assertEqual(
            Mail.get_config_options("mailgun")["from"], "no-reply@masonite.com"
        )

    def test_view_facade(self):
        View.add_location("tests/integrations/templates")
        self.assertEqual("test", View.render("test", {"test": "test"}).get_content())



import pytest

from tests import TestCase
from src.masonite.mail import Mailable


class Welcome(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )


@pytest.mark.integrations
class TestMailgunDriver(TestCase):
    def test_send_mailable(self):
        response = (
            self.application.make("mail")
            .mailable(
                Welcome().attach("invoice", "tests/integrations/storage/invoice.pdf")
            )
            .send(driver="mailgun")
        )
        self.assertEqual(response.status_code, 200)
        # because domain is not configured we got this funny message !
        self.assertEqual("Mailgun Magnificent API", response.content.decode("utf-8"))



from src.masonite.configuration import config
from src.masonite.helpers import url, optional
from src.masonite.sessions import old
from tests import TestCase
from jinja2 import TemplateNotFound


class TestView(TestCase):
    def setUp(self):
        super().setUp()
        # keep this to have a "fresh view instance" for each test
        self.view = self.application.make("view")
        self.view.loaders = []
        self.view.composers = {}
        self.view.add_location("tests/integrations/templates")

    def test_can_pass_dict(self):
        self.assertIn("test", self.view.render("test", {"test": "test"}).get_content())

    def test_view_exists(self):
        self.assertTrue(self.view.exists("welcome"))
        self.assertFalse(self.view.exists("not_available"))

    def test_view_render_does_not_keep_previous_variables(self):
        self.view.render("test", {"var1": "var1"})
        self.view.render("test", {"var2": "var2"})

        self.assertNotIn("var1", self.view.dictionary)
        self.assertIn("var2", self.view.dictionary)

    def test_global_view_exists(self):
        self.assertTrue(self.view.exists("/tests/integrations/templates/welcome"))
        self.assertFalse(
            self.view.exists("/tests/integrations/templates/not_available")
        )

    def test_exception_view_not_found(self):
        with self.assertRaises(TemplateNotFound) as context:
            self.view.render("not_found")

        self.assertEqual(context.exception.message, "Template 'not_found' not found")

    def test_exception_include_not_found(self):
        with self.assertRaises(TemplateNotFound) as context:
            self.view.render("include")

        self.assertEqual(
            context.exception.message,
            "One of the included templates in the 'include' view could not be found",
        )

    def test_view_gets_global_template(self):
        self.assertEqual(
            self.view.render(
                "/tests/integrations/templates/test", {"test": "test"}
            ).get_content(),
            "test",
        )

    def test_view_extends_without_dictionary_parameters(self):
        self.view.share({"test": "test"})
        self.assertEqual(self.view.render("test").get_content(), "test")

    def test_composers(self):
        view = self.view.composer("test", {"test": "test"})

        self.assertEqual(view.composers, {"test": {"test": "test"}})
        self.assertEqual(view.render("test").rendered_template, "test")

    def test_composers_load_all_views_with_asterisks(self):

        self.view.composer("*", {"test": "test"})

        self.assertEqual(self.view.composers, {"*": {"test": "test"}})

        self.assertEqual(self.view.render("test").get_content(), "test")

    def test_composers_with_wildcard_base_view(self):
        self.view.composer("mail*", {"to": "test_user"})

        self.assertEqual(self.view.composers, {"mail*": {"to": "test_user"}})

        self.assertIn("test_user", self.view.render("mail/welcome").get_content())

    def test_composers_with_wildcard_base_view_route(self):
        self.view.composer("mail*", {"to": "test_user"})

        self.assertEqual(self.view.composers, {"mail*": {"to": "test_user"}})

        self.assertIn("test_user", self.view.render("mail/welcome").get_content())

    def test_composers_with_wildcard_lower_directory_view_and_incorrect_shortend_wildcard(
        self,
    ):
        self.view.composer("mail/wel*", {"to": "test_user"})

        self.assertEqual(self.view.composers, {"mail/wel*": {"to": "test_user"}})

        assert "test_user" not in self.view.render("mail/welcome").get_content()

    def test_composers_load_all_views_with_list(self):
        self.view.composer(["home", "test"], {"test": "test"})

        self.assertEqual(
            self.view.composers, {"home": {"test": "test"}, "test": {"test": "test"}}
        )

        self.assertEqual(self.view.render("test").rendered_template, "test")

    def test_view_share_updates_dictionary_not_overwrite(self):
        self.view.share({"test1": "test1"})
        self.view.share({"test2": "test2"})

        self.assertEqual(self.view._shared["test1"], "test1")
        self.assertEqual(self.view._shared["test2"], "test2")
        self.view.render("test", {"var1": "var1"})
        self.assertIn("test1", self.view.dictionary.keys())
        self.assertIn("test2", self.view.dictionary.keys())
        self.assertIn("var1", self.view.dictionary.keys())

    def test_can_use_namespaced_view(self):
        self.view.add_namespaced_location("auth", "tests/integrations/templates/auth")
        self.assertIn("Welcome", self.view.render("auth:home").get_content())

    def test_can_access_shared_helpers(self):
        class SomeObject:
            value = "object_value"

        obj = SomeObject()
        self.application.make("session").flash("username", "Sam")

        content = self.view.render("test_helpers", {"obj": obj}).get_content()
        self.assertIn(
            config("application.app_url"),
            content,
        )
        self.assertIn(url.asset("local", "avatar.jpg"), content)
        self.assertIn(url.url("welcome"), content)
        self.assertIn(old("username"), content)
        self.assertIn(optional(obj).value, content)
        self.assertIn("auth/base template exists", content)
        self.assertIn("name='__method' value='PATCH'", content)



from typing import Any

class Session:
    """Session facade."""

    def add_driver(name: str, driver: Any) -> None:
        """Register a new session driver with the given name."""
        ...
    def driver(driver: str) -> Any:
        """Get a registered session driver with the given name."""
        ...
    def set_configuration(config: dict) -> "Session":
        """Set session driver options."""
        ...
    def get_driver(name: str = None) -> Any:
        """Get the default session driver or the driver with the given name."""
        ...
    def get_config_options(driver: str = None) -> dict:
        """Get the options of the default session driver or of the driver with the given name."""
        ...
    def start(driver: str = None) -> "Session":
        """Initialize session."""
        ...
    def get_data(self) -> dict:
        """Get all session data."""
        ...
    def save(driver: str = None) -> None:
        """Save session data for the default session driver or the given named driver."""
        ...
    def set(key: str, value: Any) -> None:
        """Save value in default session."""
        ...
    def increment(key: str, count: int = 1) -> None:
        """Increment session key with given count."""
        ...
    def decrement(key: str, count: int = 1) -> None:
        """Decrement session key with given count."""
        ...
    def has(key: str) -> bool:
        """Check if key is present in default session."""
        ...
    def get(key: str) -> Any:
        """Get value of the given key in default session."""
        ...
    def pull(key: str) -> Any:
        """Get and remove value for the given key in session."""
        ...
    def flush(self) -> None:
        """Delete all keys from session."""
        ...
    def delete(key: str) -> "None|Any":
        """Delete the given key from session."""
        ...
    def flash(key: str, value: Any) -> None:
        """Save temporary value into session."""
        ...
    def all(self) -> dict:
        """Get all session data."""
        ...



"""New Middleware Command."""
import inflection
import os

from ..utils.location import base_path
from ..utils.filesystem import get_module_dir, make_directory, render_stub_file
from .Command import Command


class MakeMiddlewareCommand(Command):
    """
    Creates a new middleware class.

    middleware
        {name : Name of the middleware}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        if not name.endswith("Middleware"):
            name += "Middleware"

        content = render_stub_file(self.get_middleware_path(), name)

        relative_filename = os.path.join(
            self.app.make("middlewares.location"), name + ".py"
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)
        if os.path.exists(relative_filename) and not self.option("force"):
            self.warning(
                f"{relative_filename} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(filepath, "w") as f:
            f.write(content)

        self.info(f"Middleware Created ({relative_filename})")

    def get_middleware_path(self):
        return os.path.join(
            get_module_dir(__file__), "../stubs/middlewares/Middleware.py"
        )



from masoniteorm.models import Model

from tests import TestCase
from src.masonite.notification import Notification, Notifiable
from src.masonite.mail import Mailable
from src.masonite.facades import Config


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class WelcomeUserNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .to(notifiable.email)
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text(f"Hello {notifiable.name}")
        )

    def via(self, notifiable):
        return ["mail"]


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
        )

    def via(self, notifiable):
        return ["mail"]


class CustomNotification(Notification, Mailable):
    def to_mail(self, notifiable):
        return (
            self.subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
            .driver("mailgun")
        )

    def via(self, notifiable):
        return ["mail"]


class TestMailDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")
        # use terminal driver as email driver for testing
        self.default_driver = Config.get("mail.drivers.default")
        Config.set("mail.drivers.default", "terminal")

    def tearDown(self):
        super().tearDown()
        Config.set("mail.drivers.default", self.default_driver)

    def test_send_to_anonymous(self):
        self.notification.route("mail", "test@mail.com").send(WelcomeNotification())

    def test_send_to_notifiable(self):
        user = User.find(1)
        user.notify(WelcomeUserNotification())

    def test_mail_driver_can_be_overriden_in_notification(self):
        mail = self.fake("mail")
        self.notification.route("mail", "test@mail.com").send(CustomNotification())
        mail.seeDriverWas("mailgun")
        self.restore("mail")



from tests import TestCase
from src.masonite.routes import Route, HTTPRoute
from src.masonite.exceptions.exceptions import RouteNotFoundException


class TestHttpRequests(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.get("/test/users/@id", "WelcomeController@show_user"),
            Route.get("/test-with-input", "WelcomeController@form_with_input"),
            Route.get("/server-error", "WelcomeController@server_error"),
        )

    def test_csrf_request(self):
        self.withoutCsrf()
        return self.post("/").assertContains("Welcome")

    def test_find_or_fail_during_request(self):
        self.get("/test/users/1").assertOk()
        self.withExceptionsHandling()
        with self.debugMode(False):
            self.get("/test/users/10").assertNotFound()

    def test_route_not_found_throw_exception_in_debug_mode(self):
        with self.debugMode():
            with self.assertRaises(RouteNotFoundException) as e:
                self.get("/unknown/route")

            self.assertEqual(str(e.exception), "GET /unknown/route : 404 Not Found")

    def test_route_not_found_with_debug_off(self):
        self.withExceptionsHandling()  # enable exception handler
        with self.debugMode(False):
            self.get("/unknown/route").assertNotFound().assertViewIs("errors/404")

    def test_server_error_with_debug_off(self):
        self.withExceptionsHandling()  # enable exception handler
        with self.debugMode(False):
            self.get("/server-error").assertError().assertViewIs("errors/500")

    def test_server_error_with_debug_on(self):
        with self.debugMode():
            with self.assertRaises(Exception):
                self.get("/server-error")

    def test_old_helper(self):
        self.get("/test-with-input", {"name": "Sam"})
        assert self.application.make("request").old("name") == "Sam"
        assert self.application.make("request").old("wrong-input") == ""

    def test_get_route(self):
        self.get("/")
        route = self.application.make("request").get_route()
        assert isinstance(route, HTTPRoute)



"""Loader class to easily list, find or load any object in a given module, or folder."""
import inspect
import pkgutil
import os

from ..exceptions import LoaderNotFound
from ..utils.str import as_filepath
from ..utils.structures import load


def parameters_filter(obj_name, obj):
    return (
        obj_name.isupper()
        and not obj_name.startswith("__")
        and not obj_name.endswith("__")
    )


class Loader:
    def get_modules(self, files_or_directories, raise_exception=False):
        if not isinstance(files_or_directories, list):
            files_or_directories = [files_or_directories]

        _modules = {}
        module_paths = list(map(as_filepath, files_or_directories))
        for module_loader, name, _ in pkgutil.iter_modules(module_paths):
            module = load(
                f"{os.path.relpath(module_loader.path)}.{name}",
                raise_exception=raise_exception,
            )
            _modules.update({name: module})
        return _modules

    def find(self, class_instance, paths, class_name, raise_exception=False):
        _classes = self.find_all(class_instance, paths)
        for name, obj in _classes.items():
            if name == class_name:
                return obj
        if raise_exception:
            raise LoaderNotFound(
                f"No {class_instance} named {class_name} has been found in {paths}"
            )
        return None

    def find_all(self, class_instance, paths, raise_exception=False):
        _classes = {}
        for module in self.get_modules(paths).values():
            for obj_name, obj in inspect.getmembers(module):
                # check if obj is the same class as the given one
                if inspect.isclass(obj) and issubclass(obj, class_instance):
                    # check if the class really belongs to those paths to load internal only
                    if obj.__module__.startswith(module.__package__):
                        _classes.update({obj_name: obj})
        if not len(_classes.keys()) and raise_exception:
            raise LoaderNotFound(f"No {class_instance} have been found in {paths}")
        return _classes

    def get_object(self, path_or_module, object_name, raise_exception=False):
        return load(path_or_module, object_name, raise_exception=raise_exception)

    def get_objects(self, path_or_module, filter_method=None, raise_exception=False):
        """Returns a dictionary of objects from the given path (file or dotted). The dictionary can
        be filtered if a given callable is given."""
        if isinstance(path_or_module, str):
            module = load(path_or_module, raise_exception=raise_exception)
        else:
            module = path_or_module
        if not module:
            return None
        return dict(inspect.getmembers(module, filter_method))

    def get_parameters(self, module_or_path):
        _parameters = {}
        for name, obj in self.get_objects(module_or_path).items():
            if parameters_filter(name, obj):
                _parameters.update({name: obj})

        return _parameters



from .BroadcastDriver import BroadcastDriver
from .DatabaseDriver import DatabaseDriver
from .MailDriver import MailDriver
from .SlackDriver import SlackDriver
from .vonage.VonageDriver import VonageDriver



from tests import TestCase
from src.masonite.notification import Sms


class Welcome(Sms):
    def build(self):
        return self.to("+33612345678").from_("+44123456789").text("Masonite 4")


class TestSms(TestCase):
    def test_build_sms(self):
        sms = Welcome().build().get_options()
        self.assertEqual(sms.get("to"), "+33612345678")
        self.assertEqual(sms.get("from"), "+44123456789")
        self.assertEqual(sms.get("text"), "Masonite 4")
        self.assertEqual(sms.get("type"), "text")

    def test_set_unicode(self):
        sms = Welcome().set_unicode().build().get_options()
        self.assertEqual(sms.get("type"), "unicode")

    def test_adding_client_ref(self):
        sms = Welcome().client_ref("ABCD").build().get_options()
        self.assertEqual(sms.get("client-ref"), "ABCD")



"""Route List Command"""
import re
from cleo import Command

from ...utils.str import get_controller_name


class RouteListCommand(Command):
    """
    List all your application routes.

    routes:list
        {--M|methods= : Filter by a list of HTTP methods: GET or GET,POST}
        {--N|name= : Filter by route name}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        # get filters
        methods = self.option("methods")
        if methods:
            methods = list(map(lambda m: m.lower(), methods.split(",")))
        name = self.option("name")

        # get routes
        router = self.app.make("router")
        routes = router.routes

        # filter by HTTP methods
        if methods:
            routes = list(
                filter(
                    lambda route: set(route.request_method).intersection(methods),
                    routes,
                )
            )
        # filter by name
        if name:
            routes = list(
                filter(
                    lambda route: re.findall(
                        name, route.get_name() or "", re.IGNORECASE
                    ),
                    routes,
                )
            )

        # build routes table
        header = ["URI", "Name", "Method(s)", "Controller", "Middleware(s)"]
        rows = list(map(lambda route: self.format_route_as_row(route), routes))
        rows.sort()
        self.render_table(header, rows)

    def format_route_as_row(self, route):
        """Format a Route object as a table row."""
        controller = get_controller_name(route.controller)
        row = [
            route.url,
            route.get_name() or "",
            "/".join(map(lambda m: m.upper(), route.request_method)),
            controller,
            ",".join(route.get_middlewares()),
        ]
        return row



from .AuthorizationResponse import AuthorizationResponse


class Policy:
    def allow(self, message="", code=None):
        return AuthorizationResponse.allow(message, code)

    def deny(self, message="", code=None):
        return AuthorizationResponse.deny(message, code)



"""Base Database Seeder Module."""

from masoniteorm.seeds import Seeder
from .user_table_seeder import UserTableSeeder


class DatabaseSeeder(Seeder):
    def run(self):
        """Run the database seeds."""
        self.call(UserTableSeeder)



from tests import TestCase
from src.masonite.facades import Config
from src.masonite.configuration import config
from src.masonite.exceptions import InvalidConfigurationSetup


PACKAGE_PARAM = 1
OTHER_PARAM = 3


class TestConfiguration(TestCase):
    def test_config_is_loaded(self):
        self.assertGreater(len(Config._config.keys()), 0)

    def test_base_configuration_files_can_be_accessed(self):
        self.assertIsNotNone("config.application")
        self.assertIsNotNone("config.auth")
        self.assertIsNotNone("config.broadcast")
        self.assertIsNotNone("config.cache")
        self.assertIsNotNone("config.database")
        self.assertIsNotNone("config.filesystem")
        self.assertIsNotNone("config.mail")
        self.assertIsNotNone("config.notification")
        self.assertIsNotNone("config.providers")
        self.assertIsNotNone("config.queue")
        self.assertIsNotNone("config.session")

    def test_config_helper(self):
        self.assertEqual(config("auth.guards.default"), "web")

    def test_config_facade(self):
        self.assertEqual(Config.get("auth.guards.default"), "web")

    def test_config_use_default_if_not_exist(self):
        self.assertEqual(config("some.app"), None)
        self.assertEqual(config("some.app", 0), 0)

    def test_config_set_value(self):
        original_value = config("cache.stores.redis.port")
        Config.set("cache.stores.redis.port", 1000)
        self.assertNotEqual(Config.get("cache.stores.redis.port"), original_value)
        self.assertEqual(Config.get("cache.stores.redis.port"), 1000)
        # reset to original value
        Config.set("cache.stores.redis.port", original_value)

    def test_can_load_non_foundation_config_in_project(self):
        self.assertEqual(config("package.package_param"), "package_value")

    def test_cannot_override_foundation_config(self):
        with self.assertRaises(InvalidConfigurationSetup):
            Config.merge_with("auth", {"key": "val"})

    def test_can_merge_external_config_with_project_config(self):
        # reset test package config for idempotent tests
        Config.set("package", {"package_param": "package_value"})

        package_default_config = {"PACKAGE_PARAM": "default_value", "OTHER_PARAM": 2}
        Config.merge_with("package", package_default_config)
        self.assertEqual(config("package.package_param"), "package_value")
        self.assertEqual(config("package.other_param"), 2)

    def test_can_merge_external_config_using_path(self):
        # reset test package config for idempotent tests
        Config.set("package", {"package_param": "package_value"})

        Config.merge_with("package", "tests/core/configuration/test_config.py")
        self.assertEqual(config("package.package_param"), "package_value")
        self.assertEqual(config("package.other_param"), 3)



from tests import TestCase
from src.masonite.foundation import Application
import os


class TestAppApplication(TestCase):
    def test_initialize_application(self):
        app = Application(os.getcwd())
        self.assertTrue(app)



from unittest.mock import MagicMock

from src.masonite.middleware import MiddlewareCapsule
from tests import TestCase


class MockMiddleware:
    def before(self, request, response, arg1):
        return request

    def after(self, request, response):

        return request


class TestMiddleware(TestCase):
    def test_can_create_capsule(self):
        capsule = MiddlewareCapsule()
        self.assertTrue(capsule)

    def test_can_add_middleware(self):
        capsule = MiddlewareCapsule()
        capsule.add({"mock": MockMiddleware})
        capsule.add([MockMiddleware])

        self.assertTrue(len(capsule.route_middleware) == 1)
        self.assertTrue(len(capsule.http_middleware) == 1)

    def test_can_add_and_remove_middleware(self):
        capsule = MiddlewareCapsule()
        capsule.add({"mock": MockMiddleware})
        capsule.add([MockMiddleware])
        capsule.remove(MockMiddleware)

        self.assertTrue(len(capsule.route_middleware) == 1)
        self.assertTrue(len(capsule.http_middleware) == 0)

    def test_can_get_multiple_middleware(self):
        capsule = MiddlewareCapsule()
        capsule.add(
            {
                "mock": MockMiddleware,
                "mock1": MockMiddleware,
                "mock2": MockMiddleware,
                "mock3": [MockMiddleware, MockMiddleware],
            }
        )
        capsule.add([MockMiddleware])
        capsule.remove(MockMiddleware)

        self.assertTrue(
            len(capsule.get_route_middleware(["mock", "mock1", "mock2"])) == 3
        )
        self.assertTrue(
            len(capsule.get_route_middleware(["mock", "mock1", "mock2", "mock3"])) == 5
        )

    def test_can_run_middleware_with_args(self):
        request = self.make_request()
        response = self.make_response()
        capsule = MiddlewareCapsule()
        MockMiddleware.before = MagicMock(return_value=request)
        capsule.add(
            {
                "mock": MockMiddleware,
            }
        )

        capsule.run_route_middleware(["mock:arg1,arg2"], request, response)
        MockMiddleware.before.assert_called_with(request, response, "arg1", "arg2")

    def test_can_use_request_inputs_as_args(self):
        # this create a request with @user_id and @id as in input
        request = self.make_request(query_string="user_id=3&id=1")
        response = self.make_response()
        capsule = MiddlewareCapsule()
        MockMiddleware.before = MagicMock(return_value=request)
        capsule.add(
            {
                "mock": MockMiddleware,
            }
        )

        capsule.run_route_middleware(["mock:@user_id,@id"], request, response)
        MockMiddleware.before.assert_called_with(request, response, "3", "1")

    def test_can_mix_args_and_request_inputs(self):
        # this create a request with @user_id as in input
        request = self.make_request(query_string="user_id=3")
        response = self.make_response()
        capsule = MiddlewareCapsule()
        MockMiddleware.before = MagicMock(return_value=request)
        capsule.add(
            {
                "mock": MockMiddleware,
            }
        )

        capsule.run_route_middleware(["mock:@user_id,value"], request, response)
        MockMiddleware.before.assert_called_with(request, response, "3", "value")



from src.masonite.auth import Sign
from src.masonite.foundation import response_handler
from src.masonite.storage import StorageCapsule
from src.masonite.environment import LoadEnvironment
from src.masonite.configuration import Configuration, config
from src.masonite.middleware import (
    VerifyCsrfToken,
    SessionMiddleware,
    ShareErrorsInSessionMiddleware,
    EncryptCookies,
    LoadUserMiddleware,
    MaintenanceModeMiddleware,
    ClearDumpsBetweenRequestsMiddleware,
    IpMiddleware,
)
from src.masonite.routes import Route
from src.masonite.utils.structures import load
from src.masonite.utils.location import base_path


class Kernel:

    http_middleware = [
        EncryptCookies,
        ClearDumpsBetweenRequestsMiddleware,
        IpMiddleware,
        MaintenanceModeMiddleware,
    ]
    route_middleware = {
        "web": [
            SessionMiddleware,
            ShareErrorsInSessionMiddleware,
            LoadUserMiddleware,
            VerifyCsrfToken,
        ]
    }

    def __init__(self, app):
        self.application = app

    def register(self):
        self.load_environment()
        self.register_configurations()
        self.register_middleware()
        self.register_routes()
        self.register_database()
        self.register_templates()
        self.register_storage()

    def load_environment(self):
        LoadEnvironment()

    def register_configurations(self):
        # load configuration
        self.application.bind("config.location", "tests/integrations/config")
        configuration = Configuration(self.application)
        configuration.load()
        self.application.bind("config", configuration)
        key = config("application.key")
        self.application.bind("key", key)
        self.application.bind("sign", Sign(key))

        # set locations
        self.application.bind("controllers.location", "tests/integrations/controllers")
        self.application.bind("jobs.location", "tests/integrations/jobs")
        self.application.bind("mailables.location", "tests/integrations/mailables")
        self.application.bind("providers.location", "tests/integrations/providers")
        self.application.bind("listeners.location", "tests/integrations/listeners")
        self.application.bind("validation.location", "tests/integrations/validation")
        self.application.bind("tasks.location", "tests/integrations/tasks")
        self.application.bind("events.location", "tests/integrations/events")
        self.application.bind("policies.location", "tests/integrations/policies")
        self.application.bind(
            "notifications.location", "tests/integrations/notifications"
        )
        self.application.bind("resources.location", "tests/integrations/resources")
        self.application.bind("models.location", "tests/integrations/app")
        self.application.bind("observers.location", "tests/integrations/app")
        self.application.bind("commands.location", "tests/integrations/commands")
        self.application.bind(
            "middlewares.location", "tests/integrations/app/middlewares"
        )

    def register_middleware(self):
        self.application.make("middleware").add(self.route_middleware).add(
            self.http_middleware
        )

    def register_routes(self):
        Route.set_controller_locations(self.application.make("controllers.location"))

        self.application.bind("routes.location", "tests/integrations/web")
        self.application.make("router").add(
            Route.group(
                load(self.application.make("routes.location"), "ROUTES", []),
                middleware=["web"],
            )
        )
        self.application.bind("routes.api.location", "tests/integrations/routes/api")

    def register_templates(self):
        self.application.bind("views.location", "tests/integrations/templates")

    def register_database(self):
        from masoniteorm.query import QueryBuilder

        self.application.bind(
            "builder",
            QueryBuilder(connection_details=config("database.databases")),
        )

        self.application.bind(
            "migrations.location", "tests/integrations/databases/migrations"
        )
        self.application.bind("seeds.location", "tests/integrations/databases/seeds")

        self.application.bind("resolver", config("database.db"))

    def register_storage(self):
        storage = StorageCapsule()
        storage.add_storage_assets(
            {
                # folder          # template alias
                "tests/integrations/storage/static": "static/",
                "tests/integrations/storage/compiled": "static/",
                "tests/integrations/storage/uploads": "static/",
                "tests/integrations/storage/public": "/",
            }
        )
        self.application.bind("storage_capsule", storage)

        self.application.set_response_handler(response_handler)
        self.application.use_storage_path(base_path("tests/integrations/storage"))



class Cookie:
    def __init__(
        self,
        name,
        value,
        expires=None,
        http_only=True,
        path="/",
        timezone=None,
        secure=False,
        samesite="Strict",
    ):
        self.name = name
        self.value = value
        self.http_only = http_only
        self.secure = secure
        self.expires = expires
        self.timezone = timezone
        self.samesite = samesite
        self.path = path

    def render(self):
        response = f"{self.name}={self.value};"
        if self.http_only:
            response += "HttpOnly;"
        if self.secure:
            response += "Secure;"
        if self.expires:
            if self.timezone:
                response += f"Expires={self.expires} {self.timezone};"
            else:
                response += f"Expires={self.expires};"
        if self.path:
            response += f"Path={self.path};"
        if self.samesite:
            response += f"SameSite={self.samesite};"

        return response



import builtins

from exceptionite import Handler
from exceptionite.options import DEFAULT_OPTIONS
from exceptionite.renderers import JSONRenderer
from exceptionite.solutions import MasoniteSolutions

from .Provider import Provider
from ..routes import Route
from ..configuration import config
from ..exceptions.ExceptionHandler import ExceptionHandler
from ..dumps import Dumper
from ..exceptions import (
    DumpExceptionHandler,
    HttpExceptionHandler,
    ModelNotFoundHandler,
)
from ..exceptions.exceptionite.controllers import ExceptioniteController
from ..exceptions.exceptionite.tabs import DumpsTab
from ..exceptions.exceptionite.blocks import RequestBlock, AppBlock, ConfigBlock
from ..exceptions.exceptionite import solutions


class ExceptionProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        exceptionite = Handler()
        options = config("exceptions")
        # @removed:5.0.0
        # old project won't have new options for exceptions
        if not options.get("options.editor"):
            options = DEFAULT_OPTIONS
        exceptionite.set_options(options)

        # configure exceptionite for Masonite specifically
        exceptionite.app = self.application
        self.application.make("router").add(
            Route.group(
                [
                    Route.post(
                        "/_exceptionite/actions/", ExceptioniteController.run_action
                    )
                ]
            )
        )
        exceptionite.add_renderer("json", JSONRenderer)
        exceptionite.renderer("web").add_tabs(DumpsTab)
        exceptionite.renderer("web").tab("context").add_blocks(
            RequestBlock, AppBlock, ConfigBlock
        )
        exceptionite.renderer("web").tab("solutions").block(
            "possible_solutions"
        ).register(
            solutions.TableNotFound(),
            solutions.MissingCSRFToken(),
            solutions.InvalidCSRFToken(),
            solutions.TemplateNotFound(),
            solutions.NoneResponse(),
            solutions.RouteMiddlewareNotFound(),
            *MasoniteSolutions.get()
        )

        exception_handler = ExceptionHandler(self.application)
        exception_handler.add_driver("exceptionite", exceptionite)
        self.application.bind("exception_handler", exception_handler)

        # dumper
        dumper = Dumper(self.application)
        self.application.bind("dumper", dumper)
        builtins.dd = dumper.dd
        builtins.dump = dumper.dump
        builtins.clear_dumps = dumper.clear
        self.application.bind(
            "DumpExceptionHandler", DumpExceptionHandler(self.application)
        )
        self.application.bind(
            "HttpExceptionHandler", HttpExceptionHandler(self.application)
        )
        self.application.bind(
            "ModelNotFoundHandler", ModelNotFoundHandler(self.application)
        )

    def boot(self):
        pass



from __future__ import annotations
import os
import sys
from typing import Callable, List, Tuple, TYPE_CHECKING, Iterator, Type

from ..container import Container
from ..facades import Config
from ..environment import env

if TYPE_CHECKING:
    from ..providers import Provider

ResponseHandler = Callable[[str, List[Tuple]], None]


class Application(Container):
    def __init__(self, base_path: str = None):
        self.base_path: str = base_path
        self.storage_path: str = None
        self.response_handler: ResponseHandler
        self.providers: list = []

    def set_response_handler(self, response_handler: ResponseHandler) -> None:
        self.response_handler = response_handler

    def get_response_handler(self) -> ResponseHandler:
        return self.response_handler

    def register_providers(self, *providers: Type["Provider"]) -> "Application":
        for provider_class in providers:
            provider = provider_class(self)
            provider.register()
        return self

    def use_storage_path(self, path: str) -> None:
        self.storage_path = path

    def get_storage_path(self) -> str:
        return self.storage_path

    def add_providers(self, *providers: Type["Provider"]) -> "Application":
        for provider_class in providers:
            provider = provider_class(self)
            provider.register()
            self.providers.append(provider)

        return self

    # @clean looks like that's not used anymore it's done in Route
    # is it okay ? should not be the application or event the container which holds this info ??
    # def set_controller_locations(self, location):
    #     self._controller_locations = location

    # def get_controller_locations(self, location):
    #     return self._controller_locations

    def get_providers(self) -> List["Provider"]:
        return self.providers

    def __call__(self, *args, **kwargs) -> Iterator:
        return self.response_handler(*args, **kwargs)

    def is_debug(self) -> bool:
        """Check if debug mode is enabled."""
        # @removed:5.0.0
        if Config.has("application.debug"):
            return bool(Config.get("application.debug"))
        else:
            return env("APP_DEBUG", True)

    def is_dev(self) -> bool:
        """Check if app is running in development mode."""
        return not self.is_running_tests() and os.getenv("APP_ENV") in [
            "development",
            "local",
        ]

    def is_production(self) -> bool:
        """Check if app is running in production mode."""
        return os.getenv("APP_ENV") == "production"

    def is_running_tests(self) -> bool:
        """Check if app is running tests."""

        return "pytest" in sys.modules

    def is_running_in_console(self) -> bool:
        """Check if application is running in console. This is useful to only run some providers
        logic when used in console. We can differenciate if the application is being served or
        if an application command is ran in console."""
        if len(sys.argv) > 1:
            return sys.argv[1] != "serve"
        return True

    def environment(self) -> str:
        """Helper to get current environment."""
        if self.is_running_tests():
            return "testing"
        else:
            return os.getenv("APP_ENV")



import os
import shutil
import zipfile
import tempfile
import requests
from io import BytesIO

from ..exceptions import (
    ProjectLimitReached,
    ProjectProviderTimeout,
    ProjectProviderHttpError,
    ProjectTargetNotEmpty,
)
from .Command import Command


class ProjectCommand(Command):
    """
    Creates a new Masonite project

    start
        {target? : Path of you Masonite project}
        {--b|--branch=False : Specify which branch from the Masonite repo you would like to install}
        {--r|--release=False : Specify which version of Masonite you would like to install}
        {--repo=MasoniteFramework/cookie-cutter : Specify from which repository you want to craft your project}
        {--p|--provider=github : Specify from which repository you want to craft your project github, gitlab }
    """

    providers = ["github", "gitlab"]
    # timeout in seconds for requests made to providers
    TIMEOUT = 20
    BRANCH = 4.0

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.api_base_url = None

    def handle(self):
        target = self.argument("target")

        if not target:
            target = "."

        branch = self.option("branch")
        version = self.option("release")
        repo = self.option("repo")
        provider = self.option("provider")

        if target == ".":
            to_dir = os.path.abspath(os.path.expanduser(target))
        else:
            to_dir = os.path.join(os.getcwd(), target)
        self.check_target_does_not_exist(to_dir)

        try:
            if repo and provider not in self.providers:
                return self.error(
                    "'provider' option must be in {}".format(",".join(self.providers))
                )

            self.set_api_provider_url_for_repo(provider, repo)

            # find cookie-cutter version to use when doing 'craft new' only
            if (
                repo == "MasoniteFramework/cookie-cutter"
                and provider == "github"
                and branch == "False"
                and version == "False"
            ):
                branch = self.BRANCH

            if branch != "False":
                branch_data = self.get_branch_provider_data(provider, branch)
                if "name" not in branch_data:
                    return self.error("Branch {0} does not exist.".format(branch))

                zipball = self.get_branch_archive_url(provider, repo, branch)
            elif version != "False":
                releases_data = self.get_releases_provider_data(provider)
                zipball = False
                for release in releases_data:
                    if "tag_name" in release and release["tag_name"].startswith(
                        "v{0}".format(version)
                    ):
                        self.info("Installing version {0}".format(release["tag_name"]))
                        self.line("")
                        zipball = self.get_release_archive_url_from_release_data(
                            provider, release
                        )
                        break
                if zipball is False:
                    return self.error("Version {0} could not be found".format(version))
            else:
                tags_data = self.get_releases_provider_data(provider)

                # try to find all releases which are not prereleases
                tags = []
                for release in tags_data:
                    if release["prerelease"] is False:
                        tag_key = "tag_name" if provider == "github" else "name"
                        tags.append(release[tag_key].replace("v", ""))

                tags = sorted(
                    tags, key=lambda v: [int(i) for i in v.split(".")], reverse=True
                )
                # get url from latest tagged version
                if not tags:
                    self.comment(
                        "No tags has been found, using latest commit on master."
                    )
                    zipball = self.get_branch_archive_url(provider, repo, "master")
                else:
                    zipball = self.get_tag_archive_url(provider, repo, tags[0])
        except ProjectLimitReached:
            raise ProjectLimitReached(
                "You have reached your hourly limit of creating new projects with {0}. Try again in 1 hour.".format(
                    provider
                )
            )
        except requests.Timeout:
            raise ProjectProviderTimeout(
                "{0} provider is not reachable, request timed out after {1} seconds".format(
                    provider, self.TIMEOUT
                )
            )
        except Exception as e:
            self.error(
                "The following error happened when crafting your project. Verify options are correct else open an issue at https://github.com/MasoniteFramework/masonite."
            )
            raise e

        success = False

        zipurl = zipball

        self.info("Crafting Application ...")

        # create a tmp directory to extract project template
        tmp_dir = tempfile.TemporaryDirectory()
        try:
            request = requests.get(zipurl)
            with zipfile.ZipFile(BytesIO(request.content)) as zfile:
                zfile.extractall(tmp_dir.name)
                extracted_path = os.path.join(
                    tmp_dir.name, zfile.infolist()[0].filename
                )
            success = True
        except Exception as e:
            self.error("An error occured when downloading {0}".format(zipurl))
            raise e

        if success:
            if target == ".":
                shutil.move(extracted_path, os.getcwd())
            else:
                shutil.move(extracted_path, to_dir)

            # remove tmp directory
            tmp_dir.cleanup()

            if target == ".":
                from_dir = os.path.join(os.getcwd(), zfile.infolist()[0].filename)

                for file in os.listdir(zfile.infolist()[0].filename):
                    shutil.move(os.path.join(from_dir, file), os.getcwd())
                os.rmdir(from_dir)

            self.info("Application Created Successfully!")
            if target == ".":
                self.info("Installing Dependencies...")
                self.call("install")
                self.info(
                    "Installed Successfully. Just Run `python craft serve` To Start Your Application."
                )
            else:
                self.info(
                    f"You now will have to go into your new {target} directory and run `project install` to complete the installation"
                )

            return

        else:
            self.comment("Could Not Create Application :(")

    def check_target_does_not_exist(self, target):
        """To avoid overwriting target directory and to avoid raw errors
        check that target directory does not exist."""
        if target == os.getcwd():
            return False

        if os.path.isdir(target):
            raise ProjectTargetNotEmpty(
                "{} already exists. You must craft a project in a new directory.".format(
                    target
                )
            )

    def set_api_provider_url_for_repo(self, provider, repo):
        if provider == "github":
            self.api_base_url = "https://api.github.com/repos/{0}".format(repo)
        elif provider == "gitlab":
            import urllib.parse

            repo_encoded_url = urllib.parse.quote(repo, safe="")
            self.api_base_url = "https://gitlab.com/api/v4/projects/{0}".format(
                repo_encoded_url
            )

    def get_branch_provider_data(self, provider, branch):
        if provider == "github":
            branch_data = self._get(
                "{0}/branches/{1}".format(self.api_base_url, branch)
            )
        elif provider == "gitlab":
            branch_data = self._get(
                "{0}/repository/branches/{1}".format(self.api_base_url, branch)
            )
        return branch_data.json()

    def get_branch_archive_url(self, provider, repo, branch):
        if provider == "github":
            return "https://github.com/{0}/archive/{1}.zip".format(repo, branch)
        elif provider == "gitlab":
            # here we can provide commit, branch name or tag
            return "{0}/repository/archive.zip?sha={1}".format(
                self.api_base_url, branch
            )

    def get_tag_archive_url(self, provider, repo, version):
        if provider == "github":
            tag_data = self._get(
                "{0}/releases/tags/v{1}".format(self.api_base_url, version)
            )
            return tag_data.json()["zipball_url"]
        elif provider == "gitlab":
            return self.get_branch_archive_url("gitlab", repo, "v" + version)

    def get_releases_provider_data(self, provider):
        if provider == "github":
            releases_data = self._get("{0}/releases".format(self.api_base_url))
        elif provider == "gitlab":
            releases_data = self._get("{0}/releases".format(self.api_base_url))
        return releases_data.json()

    def get_release_archive_url_from_release_data(self, provider, release):
        if provider == "github":
            return release["zipball_url"]
        elif provider == "gitlab":
            return [x for x in release["assets"]["sources"] if x["format"] == "zip"][0][
                "url"
            ]
            # could also do
            # return "{0}/repository/archive.zip?sha={1}.zip".format(self.api_base_url, branch)

    def get_tags_provider_data(self, provider):
        if provider == "github":
            releases_data = self._get("{0}/releases".format(self.api_base_url))
        elif provider == "gitlab":
            releases_data = self._get("{0}/repository/tags".format(self.api_base_url))
        return releases_data.json()

    def _get(self, request):
        data = requests.get(request, timeout=self.TIMEOUT)
        if data.status_code != 200:
            if data.reason == "rate limit exceeded":
                raise ProjectLimitReached()
            else:
                raise ProjectProviderHttpError(
                    "{0}({1}) at {2}".format(data.reason, data.status_code, data.url)
                )
        return data



import unittest
from src.masonite.validation import MessageBag


class TestMessageBag(unittest.TestCase):
    def setUp(self):
        self.errors = MessageBag.view_helper(
            {"email": ["email is required", "email must be a valid email"]}
        )

    def test_get_errors(self):
        self.assertTrue(self.errors.any())

    def test_get_messages(self):
        self.assertIn("email is required", self.errors.messages())
        self.assertIn("email must be a valid email", self.errors.messages())



import binascii
import os
import json
import io

from src.masonite.input import InputBag
from src.masonite.tests import MockInput
from tests import TestCase


def encode_multipart_formdata(fields):
    boundary = binascii.hexlify(os.urandom(16)).decode("ascii")

    body = (
        "".join(
            "--%s\r\n"
            'Content-Disposition: form-data; name="%s"\r\n'
            "\r\n"
            "%s\r\n" % (boundary, field, value)
            for field, value in fields.items()
        )
        + "--%s--\r\n" % boundary
    )

    content_type = "multipart/form-data; boundary=%s" % boundary

    return body, content_type


class TestInput(TestCase):
    def setUp(self):
        super().setUp()
        self.post_data = MockInput(
            '{"param": "hey", "foo": [9, 8, 7, 6], "bar": "baz"}'
        )
        self.bytes_data = MockInput(b"jack=Daniels")

    def test_can_parse_query_string(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "hello=you&goodbye=me"})
        self.assertEqual(bag.get("hello"), "you")
        self.assertEqual(bag.get("goodbye"), "me")

    def test_can_parse_post_data(self):
        bag = InputBag()
        bag.load(
            {
                "CONTENT_LENGTH": len(str(json.dumps({"__token": 1}))),
                "wsgi.input": io.BytesIO(bytes(json.dumps({"__token": 1}), "utf-8")),
            }
        )
        self.assertEqual(bag.get("__token"), 1)

    def test_can_parse_duplicate_values(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "filter[name]=Joe&filter[last]=Bill"})
        """
            {"filter": [{}]}
        """
        self.assertTrue("name" in bag.get("filter"))
        self.assertTrue("last" in bag.get("filter"))

    def test_all_with_values(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "hello=you"})
        """
            {"filter": [{}]}
        """
        self.assertEqual(bag.all_as_values(), {"hello": "you"})

    def test_can_get_defaults(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": ""})
        self.assertEqual(bag.get("hello", "default"), "default")
        self.assertEqual(
            bag.get("hello"), None
        )  # TODO: This should probably return a blank string instead of None
        self.assertEqual(bag.get("hello[]"), [])

    def test_all_without_internal_values(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "hello=you&__token=tok"})
        """
            {"filter": [{}]}
        """
        self.assertEqual(bag.all_as_values(internal_variables=False), {"hello": "you"})

    def test_has(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "hello=you&goodbye=me"})
        self.assertTrue(bag.has("hello", "goodbye"))

    def test_only(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "hello=you&goodbye=me&name=Joe"})
        self.assertEqual(bag.only("hello", "name"), {"hello": "you", "name": "Joe"})

    def test_only_array_based_inputs(self):
        bag = InputBag()
        bag.load({"QUERY_STRING": "user[]=user1&user[]=user2"})
        self.assertEqual(bag.get("user[]"), ["user1", "user2"])
        bag = InputBag()
        bag.load({"QUERY_STRING": "user[user1]=value&user[user2]=value"})
        self.assertEqual(bag.get("user"), {"user1": "value", "user2": "value"})

    def test_can_parse_text_plain_content_type(self):
        post_data = MockInput(
            '{"param": "hey", "foo": [9, 8, 7, 6], "bar": "baz"}'.encode("utf-8")
        )
        bag = InputBag()
        bag.load({"wsgi.input": post_data, "CONTENT_TYPE": "text/plain"})
        self.assertEqual(bag.get("param"), "hey")

    def test_can_parse_application_json_content_type(self):
        bag = InputBag()
        bag.load({"wsgi.input": self.post_data, "CONTENT_TYPE": "application/json"})
        self.assertEqual(bag.get("param"), "hey")

    def test_can_parse_form_urlencoded_content_type(self):
        bag = InputBag()
        bag.load(
            {
                "wsgi.input": self.bytes_data,
                "CONTENT_TYPE": "application/x-www-form-urlencoded",
            }
        )
        self.assertEqual(bag.get("jack"), "Daniels")

    def test_can_parse_multipart_formdata_content_type(self):
        data, content_type = encode_multipart_formdata({"key": "value", "test": 1})
        bag = InputBag()
        bag.load(
            {
                "REQUEST_METHOD": "POST",
                "CONTENT_TYPE": content_type,
                "CONTENT_LENGTH": str(len(data.encode("utf-8"))),
                "wsgi.input": io.BytesIO(data.encode("utf-8")),
            }
        )

        self.assertEqual(bag.get("key"), "value")
        self.assertEqual(bag.get("test"), "1")

    def test_advanced_dict_parse(self):
        bag = InputBag()
        inputs = bag.parse_dict(
            {"user[][name]": ["Joe"], "user[][email]": ["joe@masoniteproject.com"]}
        )
        self.assertEqual(
            inputs, {"user": [{"name": "Joe"}, {"email": "joe@masoniteproject.com"}]}
        )
        inputs = bag.parse_dict(
            {"user[name]": ["Joe"], "user[email]": ["joe@masoniteproject.com"]}
        )
        self.assertEqual(
            inputs, {"user": {"email": "joe@masoniteproject.com", "name": "Joe"}}
        )

    def test_can_parse_nested_post_data(self):
        # application/json
        bag = InputBag()
        data = {"key": "val", "a": {"b": {"c": 1}}}
        bag.load(
            {
                "CONTENT_TYPE": "application/json",
                "CONTENT_LENGTH": len(str(json.dumps(data))),
                "wsgi.input": io.BytesIO(bytes(json.dumps(data), "utf-8")),
            }
        )
        self.assertEqual(bag.get("key"), "val")
        self.assertEqual(bag.get("a.b.c"), 1)



"""UserTableSeeder Seeder."""

from masoniteorm.seeds import Seeder
from tests.integrations.app.User import User


class UserTableSeeder(Seeder):
    def run(self):
        """Run the database seeds."""
        User.create(
            {
                "name": "idmann509",
                "email": "idmann509@gmail.com",
                "password": "secret",
                "phone": "+123456789",
            }
        )



from .. import Middleware
from ...facades import Auth


class LoadUserMiddleware(Middleware):
    def before(self, request, _):
        request.set_user(Auth.user())
        return request

    def after(self, request, _):
        return request



from masoniteorm.migrations import Migration


class CreateNotificationsTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("notifications") as table:
            table.big_increments("id").primary()
            table.string("type")
            table.text("data")
            table.morphs("notifiable")
            table.datetime("read_at").nullable()
            table.timestamps()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("notifications")



import jwt
import pendulum
from ..routes import Route
from .controllers import AuthenticationController
from ..utils.str import random_string


class Api:
    def __init__(self, application, driver_config=None):
        self.application = application

    def set_configuration(self, config):
        self.config = config
        return self

    def generate_token(self):
        secret = self.config.get("jwt").get("secret")
        algorithm = self.config.get("jwt").get("algorithm")
        expire_minutes = self.config.get("jwt").get("expires")
        version = self.config.get("jwt").get("version")
        if expire_minutes:
            expire_minutes = (
                pendulum.now(tz="UTC").add(minutes=expire_minutes).to_datetime_string()
            )
        token = jwt.encode(
            {"expires": expire_minutes, "version": version, "random": random_string(10)}, secret, algorithm=algorithm
        )

        return token

    def get_token(self):
        request = self.application.make("request")
        token = request.input("token")

        if token:
            return token

        header = request.header("Authorization")

        if header:
            return header.replace("Bearer ", "")

    def validate_token(self, token):
        secret = self.config.get("jwt").get("secret")
        algorithm = self.config.get("jwt").get("algorithm")
        expire_minutes = self.config.get("jwt").get("expires")
        authenticates = self.config.get("jwt").get("authenticates")
        version = self.config.get("jwt").get("version")
        if expire_minutes:
            expire_minutes = (
                pendulum.now(tz="UTC").add(minutes=expire_minutes).to_datetime_string()
            )

        try:
            unencrypted_token = jwt.decode(token, secret, algorithms=[algorithm])
        except (jwt.InvalidSignatureError, jwt.DecodeError):
            return False
        expires = unencrypted_token.get("expires")

        if version:
            if unencrypted_token["version"] != version:
                return False

        if authenticates:
            return self.attempt_by_token(token)

        if not expires:
            return True

        expired = pendulum.parse(expires, tz="UTC").is_past()

        if expired:
            return False

        return True

    def regenerate_token(self, token):
        # if the token can be decoded, regenerate new token
        secret = self.config.get("jwt").get("secret")
        algorithm = self.config.get("jwt").get("algorithm")
        try:
            jwt.decode(token, secret, algorithms=[algorithm])
            return self.generate_token()
        except jwt.DecodeError:
            pass

        return False

    def attempt_by_token(self, token):
        model = self.config.get("jwt").get("model")()
        return model.attempt_by_token(token)

    @classmethod
    def routes(cls, auth_route="/api/auth", reauth_route="/api/reauth"):
        return [
            Route.post("/api/auth", AuthenticationController.auth),
            Route.post("/api/reauth", AuthenticationController.reauth),
        ]



import pytest
from tests import TestCase
from src.masonite.notification import Notification, Notifiable
from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]

    def route_notification_for_broadcast(self):
        return f"user.{self.id}"


class WelcomeNotification(Notification):
    def to_broadcast(self, notifiable):
        return {"data": "Welcome"}

    def via(self, notifiable):
        return ["broadcast"]


@pytest.mark.integrations
class TestBroadcastDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    def test_send_to_anonymous(self):
        self.notification.route("broadcast", "all").send(WelcomeNotification())

    def test_send_to_notifiable(self):
        user = User.find(1)
        user.notify(WelcomeNotification())



from src.masonite.essentials.middleware import HashIDMiddleware
from src.masonite.essentials.helpers.hashid import hashid
from tests import TestCase


class TestHashID(TestCase):
    def test_hashid_hashes_integer(self):
        assert hashid(10) == "l9avmeG"

    def test_hashid_hashes_several_integers(self):
        assert hashid(10, 20, 30) == "dB1I1uo"

    def test_hashid_decodes_several_integers(self):
        assert hashid("B1I1uo", decode=True) == (10, 20, 30)

    def test_hashid_decodes_non_encoded_value_is_falsey(self):
        assert not hashid("B8I6ub", decode=True)

    def test_hashid_can_decode_dictionary(self):
        assert (
            hashid(
                {
                    "id": "l9avmeG",
                    "name": "Joe",
                },
                decode=True,
            )
            == {"id": 10, "name": "Joe"}
        )

    def test_middleware(self):
        request = self.make_request(query_string="id=l9avmeG&name=Joe")
        HashIDMiddleware().before(request, None)
        assert request.all() == {"id": 10, "name": "Joe"}



"""React Preset"""
import shutil
import os

from .Preset import Preset
from ..utils.filesystem import make_directory
from ..utils.location import resources_path, views_path


class React(Preset):
    """
    Configure the front-end scaffolding for the application to use ReactJS

    Will also remove Vue as Vue and React are a bit mutally exclusive
    """

    key = "react"
    packages = {
        "react": "^17.0.2",
        "react-dom": "^17.0.2",
        "@babel/preset-react": "^7.16.5",
    }
    removed_packages = ["vue", "vue-loader"]

    def install(self):
        """Install the preset"""
        self.update_packages(dev=True)
        self.update_webpack_mix()
        self.update_js()
        self.add_components()
        self.update_css()
        self.create_view()
        self.remove_node_modules()

    def add_components(self):
        """Copy example React component into application (delete example Vue component
        if it exists)"""
        # make components directory if does not exists
        make_directory(resources_path("js/components/Example.js"))

        # delete Vue components if exists
        vue_files = [
            resources_path("js/components/HelloWorld.vue"),
            resources_path("js/App.vue"),
        ]
        for vue_file in vue_files:
            if os.path.exists(vue_file):
                os.remove(vue_file)

        # add Vue components
        shutil.copyfile(
            self.get_template_path("Example.js"),
            resources_path("js/components/Example.js"),
        )

    def create_view(self):
        """Copy an example app view with assets included."""
        shutil.copyfile(
            self.get_template_path("app.html"), views_path("app_react.html")
        )



from src.masonite.foundation import Application, Kernel
from src.masonite.utils.location import base_path
from tests.integrations.config.providers import PROVIDERS
from tests.integrations.app.Kernel import Kernel as ApplicationKernel


# here the project base path is tests/integrations
application = Application(base_path("tests/integrations"))

"""First Bind important providers needed to start the server
"""

application.register_providers(
    Kernel,
    ApplicationKernel,
)

application.add_providers(*PROVIDERS)



from tests import TestCase

from src.masonite.routes import Route
from src.masonite.facades import Dump


class TestDumpExceptionHandler(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.get("/dd", "WelcomeController@dd"),
        )

    def tearDown(self):
        super().tearDown()
        Dump.clear()

    def test_dd_display_correct_view(self):
        self.withExceptionsHandling()
        self.get("/dd").assertOk().assertViewHas("dumps").assertContains(
            "dump and die : 2 dumps"
        ).assertContains("request").assertContains("test")

        self.assertConsoleOutputContains(">>> DUMP")
        self.assertConsoleOutputContains("test")



from ...providers import Provider
from ..helpers.hashid import hashid


class HashIDProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        self.application.make("view").share({"hashid": hashid})

    def boot(self):
        pass



from src.masonite.routes import Route
from src.masonite.broadcasting import Broadcast
from src.masonite.authentication import Auth

ROUTES = [
    Route.get("/", "WelcomeController@show").name("welcome"),
    Route.get("/flash_data", "WelcomeController@flash_data"),
    Route.get("/sessions", "WelcomeController@play_with_session").name(
        "play_with_session"
    ),
    Route.post("/", "WelcomeController@show"),
    Route.get("/contact", "WelcomeController@contact"),
    Route.post("/contact", "WelcomeController@contact_post"),
    Route.post("/input", "WelcomeController@input"),
    Route.post("/upload", "WelcomeController@upload").name("upload"),
    Route.get("/test", "WelcomeController@test"),
    Route.get("/emit", "WelcomeController@emit"),
    Route.get("/view", "WelcomeController@view"),
    Route.get("/mail", "MailableController@view"),
    Route.get("/users/@id", "WelcomeController@show_user").name("users.profile"),
]

Broadcast.routes()
Auth.routes()



from masoniteorm.models import Model
from src.masonite.authentication import Authenticates
from src.masonite.authorization import Authorizes
from src.masonite.notification import Notifiable


class User(Model, Authenticates, Authorizes, Notifiable):
    __fillable__ = ["name", "password", "email", "phone"]



import json


class MemcacheDriver:
    def __init__(self, application):
        self.application = application
        self.connection = None

    def set_options(self, options):
        self.options = options
        return self

    def get_connection(self):
        try:
            from pymemcache.client.base import Client
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'pymemcache' library. Run 'pip install pymemcache' to fix this."
            )

        if not self.connection:
            if str(self.options.get("port")) != "0":
                self.connection = Client(
                    f"{self.options.get('host')}:{self.options.get('port')}"
                )
            else:
                self.connection = Client(f"{self.options.get('host')}")

        return self.connection

    def add(self, key, value):
        if self.has(key):
            return self.get(key)

        self.put(key, value)
        return value

    def get(self, key, default=None, **options):
        if not self.has(key):
            return default

        return self.get_value(
            self.get_connection().get(f"{self.get_name()}_cache_{key}")
        )

    def put(self, key, value, seconds=0, **options):
        if isinstance(value, (dict, list)):
            value = json.dumps(value)

        return self.get_connection().set(
            f"{self.get_name()}_cache_{key}", value, expire=seconds
        )

    def has(self, key):
        return self.get_connection().get(f"{self.get_name()}_cache_{key}")

    def increment(self, key, amount=1):
        return self.put(key, str(int(self.get(key)) + amount))

    def decrement(self, key, amount=1):
        return self.put(key, str(int(self.get(key)) - amount))

    def remember(self, key, callable):
        value = self.get(key)

        if value:
            return value

        callable(self)

    def forget(self, key):
        return self.get_connection().delete(f"{self.get_name()}_cache_{key}")

    def flush(self):
        return self.get_connection().flush_all()

    def get_name(self):
        return self.options.get("name")

    def get_value(self, value):
        if isinstance(value, bytes):
            value = value.decode("utf-8")

        value = str(value)
        if value.isdigit():
            return str(value)
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value



from multiprocessing import AuthenticationError
from src.masonite.controllers import Controller
from src.masonite.exceptions.exceptions import AuthorizationException
from src.masonite.views import View
from src.masonite.response.response import Response
from src.masonite.request import Request
from src.masonite.filesystem import Storage
from src.masonite.broadcasting import Broadcast, Channel
from src.masonite.facades import Session, Config, Gate, Dump
from src.masonite.validation import Validator

from tests.integrations.app.User import User


class CanBroadcast:
    def broadcast_on(self):
        return Channel(f"private-shipped")

    def broadcast_with(self):
        return vars(self)

    def broadcast_as(self):
        return self.__class__.__name__


class OrderProcessed(CanBroadcast):
    def __init__(self):
        self.order_id = 1


class WelcomeController(Controller):
    def play_with_session(self, request: Request, view: View):
        # Session.flash("test", "hello flashed")
        # Session.set("test_persisted", "hello persisted in session")
        # request.app.make("session").set("test_persisted", "hello persisted")
        return view.render("welcome")

    def show(self, request: Request, view: View):
        request.app.make("session").flash("test", "value")
        return view.render("welcome")

    def contact(self, view: View):
        return view.render("contact")

    def contact_post(
        self, request: Request, view: View, validator: Validator, response: Response
    ):
        errors = request.validate(validator.required(["name", "email"]))
        if errors:
            return response.back().with_errors(errors)

        return view.render("contact")

    def flash_data(self, request: Request, response: Response, view: View):
        request.app.make("session").flash("test", "value")
        return response.with_input().redirect("/sessions")

    def form_with_input(self, request: Request, response: Response, view: View):
        return response.redirect("/sessions").with_input()

    def test(self):
        return "test"

    def api(self):
        return {"key": "value"}

    def emit(self, broadcast: Broadcast):
        broadcast.channel("private-orders", OrderProcessed())
        return "emitted"

    def view(self, view: View):
        return view.render("welcome")

    def upload(self, request: Request, storage: Storage):
        return storage.disk("s3").store(request.input("profile"))

    def create(self):
        return "user created", 201

    def not_found(self):
        return "not found", 404

    def unauthorized(self):
        return "unauthorized", 401

    def input(self, request: Request):
        return request.all()
        return "input"

    def forbidden(self):
        return "forbidden", 403

    def empty(self):
        return "", 204

    def redirect_url(self, response: Response):
        return response.redirect("/")

    def redirect_route(self, response: Response):
        return response.redirect(name="test")

    def redirect_route_params(self, response: Response):
        return response.redirect(name="test_params", params={"id": 1})

    def response_with_headers(self, response: Response):
        response.header("TEST", "value")
        response.header("TEST2", "value2")
        return ""

    def view_with_context(self, view: View):
        return view.render(
            "welcome",
            {"count": 1, "users": ["John", "Joe"], "other_key": {"nested": 1}},
        )

    def json(self, response: Response):
        return response.json(
            {
                "key": "value",
                "key2": [1, 2],
                "other_key": {
                    "nested": 1,
                    "nested_again": {"a": 1, "b": 2},
                },
            }
        )

    def session(self, request: Request):
        request.app.make("session").flash("key", "value")
        return "session"

    def session_with_errors(self, request: Request, response: Response):
        request.app.make("session").flash("key", "value")
        request.app.make("session").flash(
            "errors",
            {"email": "Email required", "password": "Password too short", "name": ""},
        )
        return "session"

    def with_errors(self, response: Response):
        return response.back().with_errors(
            {"email": "Email required", "password": "Password too short", "name": ""}
        )

    def session2(self, request: Request):
        request.app.make("session").flash(
            "key", {"nested": 1, "nested_again": {"key2": "value2"}}
        )
        return "session2"

    def with_params(self):
        return ""

    def auth(self, request: Request):
        request.app.make("auth").guard("web").attempt("idmann509@gmail.com", "secret")
        return "logged in"

    def not_authorized(self):
        # if current user not authorized it will raise an exception
        Gate.authorize("display-admin")

    def use_authorization_helper(self, request: Request):
        request.authorize("display-admin")
        return ""

    def authorizations(self, view: View, request: Request):
        return view.render("authorizations")

    def show_user(self, request: Request, view: View):
        user = User.find_or_fail(request.param("id"))
        return view.render("welcome")

    def dd(self, request: Request):
        dump({"test": "value"})
        dd(request)
        return ""

    def server_error(self, view: View):
        raise Exception("unknown error")

    def __call__(self):
        return "welcome"



from tests import TestCase
import os
import time
from src.masonite.broadcasting import Channel, PrivateChannel
import pytest


class CanBroadcast:
    def broadcast_on(self):
        return Channel(f"order.{self.order_id}")

    def broadcast_with(self):
        return vars(self)

    def broadcast_as(self):
        return self.__class__.__name__


class OrderProcessed(CanBroadcast):
    def __init__(self):
        self.order_id = 1


@pytest.mark.integrations
class TestFileCache(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("cache")
        self.driver = self.application.make("broadcast")

    def test_can_get_file_driver(self):
        print(self.driver.channel("order.1", "status", {"status": "processed"}))

    def test_can_fire_class(self):
        print(self.driver.channel(OrderProcessed()))



"""Vue 3 Preset"""
import shutil
import os

from .Preset import Preset
from ..utils.filesystem import make_directory
from ..utils.location import resources_path, views_path


class Vue(Preset):
    """
    Configure the front-end scaffolding for the application to use VueJS 3.0

    Will also remove React as React and Vue are a bit mutally exclusive
    """

    key = "vue"
    packages = {"vue": "^3.2.26", "vue-loader": "^16.8.3"}
    removed_packages = ["@babel/preset-react", "react", "react-dom"]

    def install(self):
        """Install the preset"""
        self.update_packages(dev=True)
        self.update_webpack_mix()
        self.update_js()
        self.add_components()
        self.update_css()
        self.create_view()
        self.remove_node_modules()

    def add_components(self):
        """Copy example VueJS component into application (delete example React component
        if it exists)"""
        # make components directory if does not exists
        make_directory(resources_path("js/components/HelloWorld.vue"))

        # delete React component if exists
        react_file = resources_path("js/components/Example.js")
        if os.path.exists(react_file):
            os.remove(react_file)

        # add Vue components
        shutil.copyfile(
            self.get_template_path("HelloWorld.vue"),
            resources_path("js/components/HelloWorld.vue"),
        )
        shutil.copyfile(self.get_template_path("App.vue"), resources_path("js/App.vue"))

    def create_view(self):
        """Copy an example app view with assets included."""
        shutil.copyfile(self.get_template_path("app.html"), views_path("app_vue3.html"))



from tests import TestCase
from src.masonite.middleware import Middleware
from src.masonite.routes import Route


class TestUpdateMiddleware(Middleware):
    def before(self, request, response):
        request.session.set("key", "value")
        request.session.flash("flash_key", "flash_value")
        return request

    def after(self, request, response):
        return request


class TestUpdateAndRedirectsMiddleware(Middleware):
    def before(self, request, response):
        request.session.set("key", "value")
        return (
            response.redirect("/home")
            .with_success("Success message")
            .with_errors("Error message")
        )

    def after(self, request, response):
        return request


class TestUpdatingSessionFromMiddlewares(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("middleware").add(
            {
                "test": TestUpdateMiddleware,
                "test_redirect": TestUpdateAndRedirectsMiddleware,
            }
        )
        self.setRoutes(
            Route.get("/", "WelcomeController@test").middleware("test"),
            Route.get("/redirect", "WelcomeController@test").middleware(
                "test_redirect"
            ),
            Route.get("/home", "WelcomeController@test"),
        )

    def test_updating_session_in_middleware_before(self):
        self.get("/").assertOk().assertSessionHas("key", "value").assertSessionHas(
            "flash_key", "flash_value"
        )

    def test_updating_session_and_redirects_in_middleware_before(self):
        response = (
            self.get("/redirect")
            .assertRedirect("/home")
            .assertSessionHas("key", "value")
            .response
        )
        # here when loading /home redirected route cookies needs to be set by the previous
        # response.
        assert response.cookie_jar.exists("f_success")
        assert response.cookie_jar.exists("f_errors")
        assert response.cookie_jar.exists("s_key")



from ..Recipient import Recipient


class TerminalDriver:
    def __init__(self, application):
        self.application = application
        self.options = {}
        self.content_type = None

    def set_options(self, options):
        self.options = options
        return self

    def send(self):
        print("-------------------------------------")
        print(f"To: {Recipient(self.options.get('to')).header()}")
        print(f"From: {Recipient(self.options.get('from')).header()}")
        print(f"Cc: {Recipient(self.options.get('cc')).header()}")
        print(f"Bcc: {Recipient(self.options.get('bcc')).header()}")
        print(f"Subject: {self.options.get('subject')}")
        if self.options.get("headers"):
            for header, value in self.options.get("headers").items():
                print(f"{header}: {value}")
        print("-------------------------------------")
        print(f"{self.options.get('html_content')}")
        if self.options.get("text_content"):
            print("-------------------------------------")
            print(f"Text Content: {self.options.get('text_content')}")
        if self.options.get("attachments"):
            print("-------------------------------------")
            for index, attachment in enumerate(self.options.get("attachments")):
                index += 1
                print(f"Attachment {index}: {attachment.alias} from {attachment.path}")



"""Starts Interactive Console Command."""
import os
import code
import sys
import pendulum

from ..environment import env
from ..configuration import config
from ..utils.collections import collect
from ..utils.structures import load, data_get
from ..utils.location import base_path, config_path, models_path
from ..helpers import optional, url
from ..facades import Loader
from .. import __version__
from .Command import Command


BANNER = """ -------------------------------
| Masonite Python \033[92m{}\033[0m Console |
 -------------------------------
Masonite Version: \033[92m{}\033[0m,
Environment: \033[92m{}\033[0m,
Debug Mode: \033[92m{}\033[0m,
This interactive console has the following things imported:
    -\033[92m app (container), \033[0m
    - Utils:\033[92m {} \033[0m
    - Models:\033[92m {} \033[0m

Type `exit()` to exit."""


class TinkerCommand(Command):
    """
    Run a python shell with the container pre-loaded.

    tinker
        {--i|ipython : Run a IPython shell}
        {--d|directory=? : Override the directory to auto-load models from}
    """

    def handle(self):
        from wsgi import application
        from masoniteorm.models import Model

        python_version = "{}.{}.{}".format(
            sys.version_info.major, sys.version_info.minor, sys.version_info.micro
        )
        models_directory = self.option("directory") or models_path()
        models = Loader.find_all(Model, models_directory)
        helpers = {
            "app": application,
            "env": env,
            "pendulum": pendulum,
            "optional": optional,
            "collect": collect,
            "url": url.url,
            "asset": url.asset,
            "route": url.route,
            "load": load,
            "config": config,
            "data_get": data_get,
            "base_path": base_path,
            "config_path": config_path,
        }
        banner = BANNER.format(
            python_version,
            __version__,
            os.getenv("APP_ENV"),
            "on" if application.is_debug() else "off",
            ", ".join(list(helpers.keys())[1:]),
            ", ".join(models.keys()),
        )
        context = {**helpers, **models}

        if self.option("ipython"):
            try:
                import IPython
            except ImportError:
                raise ModuleNotFoundError(
                    "Could not find the 'IPython' library. Run 'pip install ipython' to fix this."
                )
            from traitlets.config import Config

            c = Config()
            c.TerminalInteractiveShell.banner1 = banner
            IPython.start_ipython(argv=[], user_ns=context, config=c)
        else:
            console = code.InteractiveConsole(context)
            try:
                import readline
            except ImportError:
                pass
            # When not using IPython, PYTHONSTARTUP is not used by default, so load any
            # scripts defined in this var at startup
            startup_file = os.environ.get("PYTHONSTARTUP")
            if startup_file:
                if os.path.isfile(startup_file):
                    with open(startup_file, "r") as f:
                        compiled_code = code.compile_command(
                            f.read(), startup_file, "exec"
                        )
                        console.runcode(compiled_code)
            console.interact(banner, exitmsg="Goodbye !")
